from __future__ import annotations

import pytest

from frplib.kinds      import (kind, constant, either, uniform,
                               symmetric, linear, geometric,
                               weighted_by, weighted_as, arbitrary,
                               integers, evenly_spaced, bin,
                               subsets, permutations_of,
                               )
from frplib.numeric    import as_real, as_nice_numeric
from frplib.quantity   import as_quantity
from frplib.symbolic   import symbol
from frplib.utils      import lmap, every
from frplib.vec_tuples import as_vec_tuple, vec_tuple


def values_of(u):
    return u.keys()

def weights_of(u):
    return list(u.values())


def test_kinds_factories():
    "Builtin kind factories"
    a = symbol('a')

    assert constant(1).values == {1}
    assert constant((2,)).values == {2}
    assert constant((2, 3)).values == {vec_tuple(2, 3)}

    assert either(0, 1).values == {0, 1}
    assert weights_of(either(0, 1, 2).weights) == pytest.approx([as_quantity('2/3'), as_quantity('1/3')])
    assert lmap(str, either(a, 2 * a, 2).values) == ['a', '2 a']

    u = uniform(1, 2, 3).weights
    assert values_of(u) == {vec_tuple(1), vec_tuple(2), vec_tuple(3)}
    assert weights_of(u) == pytest.approx([as_quantity('1/3'), as_quantity('1/3'), as_quantity('1/3')])

    w = weighted_as(1, 2, 3, weights=[1, 2, 4]).weights
    assert values_of(w) == {vec_tuple(1), vec_tuple(2), vec_tuple(3)}
    assert weights_of(w) == pytest.approx([as_quantity('1/7'), as_quantity('2/7'), as_quantity('4/7')])

    w = weighted_as(1, 2, 3, weights=[a, 2 * a, 4 * a]).weights
    assert values_of(w) == {vec_tuple(1), vec_tuple(2), vec_tuple(3)}
    assert weights_of(w) == pytest.approx([as_quantity('1/7'), as_quantity('2/7'), as_quantity('4/7')])

    w = weighted_as(1, 2, 3, weights=[1, 2 * a, 4 * a]).weights
    assert values_of(w) == {vec_tuple(1), vec_tuple(2), vec_tuple(3)}
    assert lmap(str, weights_of(w)) == ['1/(1 + 6 a)', '2 a/(1 + 6 a)', '4 a/(1 + 6 a)']

    w = weighted_as(a, 2 * a, 3 * a, weights=[1, 2, 4]).weights
    assert lmap(str, values_of(w)) == ['<a>', '<2 a>', '<3 a>']
    assert weights_of(w) == pytest.approx([as_quantity('1/7'), as_quantity('2/7'), as_quantity('4/7')])



from __future__ import annotations

import inspect
import math
import re
import textwrap

from collections.abc   import Iterable
from functools         import wraps
from operator          import itemgetter
from typing            import Callable, Literal, Optional, overload
from typing_extensions import Self, TypeGuard

from frplib.exceptions import OperationError, StatisticError
from frplib.numeric    import as_real, as_numeric
from frplib.protocols  import Projection, Transformable
from frplib.utils      import ensure_tuple, identity, is_interactive, is_tuple, scalarize
from frplib.vec_tuples import VecTuple, as_scalar, as_scalar_strict, as_vec_tuple

# ATTN: conversion with as_real etc in truediv, pow to prevent accidental float conversion
# This could be mitigated by eliminating ints from as_numeric*, but we'll see how this
# goes.


#
# Helpers
#


def compose2(after: 'Statistic', before: 'Statistic') -> 'Statistic':
    if after.dim == 0 or before.codim is None or after.dim == before.codim:
        def composed(*x):
            return after(before(*x))
        return Statistic(composed, dim=before.dim, codim=after.codim, name=f'{after.name}({before.name}(__))')
    raise OperationError(f'Statistics {after.name} and {before.name} are not compatible for composition.')


#
# Decorator/Wrapper to make functions auto-uncurry
#

def tuple_safe(fn: Callable, arity: Optional[int] = None) -> Callable:
    """Returns a function that can accept a single tuple or multiple individual arguments.

    Ensures that the returned function has an `arity` attribute set
    to the supplied or computed arity.
    """
    if arity is None:
        arity = len([param for param in inspect.signature(fn).parameters.values()
                     if param.kind <= inspect.Parameter.POSITIONAL_OR_KEYWORD])
    if arity == 0:
        @wraps(fn)
        def f(*x):
            if len(x) == 1 and is_tuple(x[0]):
                return as_vec_tuple(fn(x[0]))
            return as_vec_tuple(fn(x))
        setattr(f, 'arity', arity)
        return f
    elif arity == 1:
        @wraps(fn)
        def g(x):
            if is_tuple(x) and len(x) == 1:
                return as_vec_tuple(fn(x[0]))
            return as_vec_tuple(fn(x))
        setattr(g, 'arity', arity)
        return g

    @wraps(fn)
    def h(*x):
        select = itemgetter(*range(arity))
        if len(x) == 1 and is_tuple(x[0]):
            return as_vec_tuple(fn(*select(x[0])))
        return as_vec_tuple(fn(*select(x)))
    setattr(h, 'arity', arity)
    return h


#
# The Statistics Interface
#

# TODO: Need to handle dimensions of return type.
#       In particular, something like Sum returns a scalar, which needs to be wrapped in a tuple at the end ?!?
#       Also want to allow operations on tuples to be component wise and implement @ as a dot product
#       So,   Id - (1, 2, 3) is a tuple and Id @ (1, 2, 3) is a scalar wrapped
#       Add: Scalar function or wrap scalar constants in operation methods below

# Idea here is to be able to do things like
# dice | Sum > 5 and d6 ** 4 ^ Sum / 4  and d6 ** 4 ^ Mean
# or even dice ^ Sum > 5

# dice | fork(Sum, Count)               # produces tuples x :-> (sum(x), len(x))
# dice | chain(Id - 7, Abs, Id <= 3)    # should reduce to lambda x: |x - 7| <= 3
#
# Built-in statistics: Sum, Mean, Max, Min, Product, Count, Dot(c1,c2,c3,...)
# [dot product, repeat last coefficient ad infinitum; this is a function that produces a Statistic],
# Id, Proj(1,...) [func that produces a statistic, equiv to @ but allows Proj(2) - Proj(1)],
# Permutation(i1,i2,i3,...),
# Square, Cube, SumOfSquares, Sqrt, Pow(exp), Exp, Ln, Abs, Sin, Cos, Tan

# NOTE: This means that we want to accept 0-1 valued functions in conditioning predicates
# as well as boolean functions. Or maybe it's OK and just works, but probably want to convert
# boolean to 0-1 in .map().  <<<---- so events become easy

# Note: codim=0 can be used to mean the same dim as input

# ATTN: Also implement things like __is__ and __in__ so we can do X ^ (__ in {0, 1, 2})

class Statistic:
    "A function that operates on the output of an FRP."
    def __init__(
            self: Self,
            fn: Callable | 'Statistic',         # Either a Statistic or a function to be turned into one
            dim: Optional[int] = None,          # Number of arguments the function takes; 0 means tuple expected
            codim: Optional[int] = None,        # Dimension of the codomain; None means don't know
            name: Optional[str] = None,         # A user-facing name for the statistic
            description: Optional[str] = None   # A description used as a __doc__ string for the Statistic
    ) -> None:
        "Note: set arity to 0 to advise that it takes the whole tuple/vector; None means try to find out"

        if isinstance(fn, Statistic):
            self.fn: Callable = fn.fn
            self.arity: int = dim if dim is not None else fn.arity
            self.codim: Optional[int] = codim if codim is not None else fn.codim
            self._name = name or fn.name
            self.__doc__: str = self.__describe__(description or fn.description or '')
            return

        f = tuple_safe(fn, dim)
        self.fn = f
        self.arity = getattr(f, 'arity')
        self.codim = codim
        self._name = name or fn.__name__ or ''
        self.__doc__ = self.__describe__(description or fn.__doc__ or '')

    def __describe__(self, description: str, returns: Optional[str] = None) -> str:
        def splitPascal(pascal: str) -> str:
            return re.sub(r'([a-z])([A-Z])', r'\1 \2', pascal)

        my_name = splitPascal(self.__class__.__name__)
        an = 'An' if re.match(r'[AEIOU]', my_name) else 'A'
        me = f'{an} {my_name} \'{self.name}\''
        that = '' if description else ' that '
        descriptor = ' that ' + (description + '. It ' if description else '')

        scalar = ''
        if not returns:
            if self.codim == 1:
                scalar = 'returns a scalar'
            elif self.codim is not None:
                scalar = f'returns a {self.codim}-tuple'
        else:
            scalar = returns

        arity = ''
        if self.arity == 0:
            arity = 'expects a tuple'
        elif self.arity > 0:
            arity = f'expects {self.arity} argument{"s" if self.arity > 1 else ""} (or a tuple of that length)'

        conj = ' and ' if scalar and arity else that if scalar else ''
        structure = f'{arity}{conj}{scalar}.'

        return f'{me}{descriptor}{structure}'

    def __str__(self) -> str:
        return self.__doc__

    def __repr__(self) -> str:
        if is_interactive():  # Needed?
            return str(self)
        # ATTN! This looks like a bug
        return super().__repr__()

    @property
    def name(self) -> str:
        return self._name

    @property
    def dim(self) -> int:
        "Returns the dimension of the statistic, with 0 meaning it accepts an arbitrary tuple."
        return self.arity

    @property
    def description(self) -> str:
        return self.__doc__

    def __call__(self, *args):
        # It is important that Statistics are not Transformable!
        if len(args) == 1:
            if isinstance(args[0], Transformable):
                return args[0].transform(self)
            if isinstance(args[0], Statistic):
                return compose2(self, args[0])
        return self.fn(*args)

    # Comparisons (macros would be nice here)

    def __eq__(self, other):
        if isinstance(other, Statistic):
            def a_eq_b(*x):
                return self(*x) == other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_eq_b(*x):
                return self(*x) == f(*x)
            label = str(other)
        else:
            def a_eq_b(*x):
                return self(*x) == other
            label = str(other)

        # Break inheritance rules here, but it makes sense!
        return Condition(a_eq_b, dim=0, name=f'{self.name}(__) == {label}')

    def __ne__(self, other):
        if isinstance(other, Statistic):
            def a_ne_b(*x):
                return self(*x) != other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_ne_b(*x):
                return self(*x) != f(*x)
            label = str(other)
        else:
            def a_ne_b(*x):
                return self(*x) != other
            label = str(other)

        # Break inheritance rules here, but it makes sense!
        return Condition(a_ne_b, dim=0, name=f'{self.name}(__) != {label}')

    ## ATTN:FIX labels for methods below, so e.g., ForEach(2*__+1) prints out nicely

    def __le__(self, other):
        if isinstance(other, Statistic):
            def a_le_b(*x):
                return self(*x) <= other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_le_b(*x):
                return self(*x) <= f(*x)
            label = str(other)
        else:
            def a_le_b(*x):
                return self(*x) <= other
            label = str(other)

        # Break inheritance rules here, but it makes sense!
        return Condition(a_le_b, dim=0, name=f'{self.name}(__) <= {label}')

    def __lt__(self, other):
        if isinstance(other, Statistic):
            def a_lt_b(*x):
                return self(*x) < other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_lt_b(*x):
                return self(*x) < f(*x)
            label = str(other)
        else:
            def a_lt_b(*x):
                return self(*x) < other
            label = str(other)

        # Break inheritance rules here, but it makes sense!
        return Condition(a_lt_b, dim=0, name=f'{self.name}(__) < {label}')

    def __ge__(self, other):
        if isinstance(other, Statistic):
            def a_ge_b(*x):
                return self(*x) >= other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_ge_b(*x):
                return self(*x) >= f(*x)
            label = str(other)
        else:
            def a_ge_b(*x):
                return self(*x) >= other
            label = str(other)

        # Break inheritance rules here, but it makes sense!
        return Condition(a_ge_b, dim=0, name=f'{self.name}(__) >= {label}')

    def __gt__(self, other):
        if isinstance(other, Statistic):
            def a_gt_b(*x):
                return self(*x) > other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_gt_b(*x):
                return self(*x) > f(*x)
            label = str(other)
        else:
            def a_gt_b(*x):
                return self(*x) > other
            label = str(other)

        # Break inheritance rules here, but it makes sense!
        return Condition(a_gt_b, dim=0, name=f'{self.name}(__) > {label}')

    # Numeric Operations (still would like macros)

    def __add__(self, other):
        if isinstance(other, Statistic):
            def a_plus_b(*x):
                return self(*x) + other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_plus_b(*x):
                return self(*x) + f(*x)
            label = str(other)
        else:
            def a_plus_b(*x):
                return self(*x) + other
            label = str(other)

        return Statistic(a_plus_b, dim=0, name=f'{self.name}(__) + {label}')

    def __radd__(self, other):
        if callable(other):   # other cannot be a Statistic in __r*__
            f = tuple_safe(other)
            def a_plus_b(*x):
                return self(*x) + f(*x)
            label = str(other)
        else:
            def a_plus_b(*x):
                return self(*x) + other
            label = str(other)

        return Statistic(a_plus_b, dim=0, name=f'{label} + {self.name}(__)')

    def __sub__(self, other):
        if isinstance(other, Statistic):
            def a_minus_b(*x):
                return self(*x) - other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_minus_b(*x):
                return self(*x) - f(*x)
            label = str(other)
        else:
            def a_minus_b(*x):
                return self(*x) - other
            label = str(other)

        return Statistic(a_minus_b, dim=0, name=f'{self.name}(__) - {label}')

    def __rsub__(self, other):
        if callable(other):   # other cannot be a Statistic in __r*__
            f = tuple_safe(other)
            def a_minus_b(*x):
                return f(*x) - self(*x)
        else:
            def a_minus_b(*x):
                return other - self(*x)

        return Statistic(a_minus_b, dim=0, name=f'{str(other)} - {self.name}(__)')

    def __mul__(self, other):
        if isinstance(other, Statistic):
            def a_times_b(*x):
                return self(*x) * other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_times_b(*x):
                return self(*x) * f(*x)
            label = str(other)
        else:
            def a_times_b(*x):
                return self(*x) * as_real(as_scalar_strict(other))  # ATTN!
            label = str(other)

        return Statistic(a_times_b, dim=0, name=f'{self.name}(__) * {label}')

    def __rmul__(self, other):
        if callable(other):   # other cannot be a Statistic in __r*__
            f = tuple_safe(other)
            def a_times_b(*x):
                return self(*x) * f(*x)
        else:
            def a_times_b(*x):
                return self(*x) * as_real(as_scalar_strict(other))

        return Statistic(a_times_b, dim=0, name=f'{str(other)} * {self.name}(__)')

    def __truediv__(self, other):
        if isinstance(other, Statistic):
            def a_div_b(*x):
                return self(*x) / other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_div_b(*x):
                return self(*x) / f(*x)
            label = str(other)
        else:
            def a_div_b(*x):
                return self(*x) / as_real(as_scalar_strict(other))
            label = str(other)

        return Statistic(a_div_b, dim=0, name=f'{self.name}(__) / {label}')

    def __rtruediv__(self, other):
        if callable(other):   # other cannot be a Statistic in __r*__
            f = tuple_safe(other)
            def a_div_b(*x):
                return f(*x) / self(*x)
        else:
            def a_div_b(*x):
                return other / as_real(as_scalar_strict(self(*x)))

        return Statistic(a_div_b, dim=0, name=f'{str(other)} / {self.name}(__)')

    def __floordiv__(self, other):
        if isinstance(other, Statistic):
            def a_div_b(*x):
                return self(*x) // other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_div_b(*x):
                return self(*x) // f(*x)
            label = str(other)
        else:
            def a_div_b(*x):
                return self(*x) // other
            label = str(other)

        return Statistic(a_div_b, dim=0, name=f'{self.name}(__) // {label}')

    def __rfloordiv__(self, other):
        if callable(other):   # other cannot be a Statistic in __r*__
            f = tuple_safe(other)
            def a_div_b(*x):
                return f(*x) // self(*x)
        else:
            def a_div_b(*x):
                return other // self(*x)

        return Statistic(a_div_b, dim=0, name=f'{str(other)} // {self.name}(__)')

    def __mod__(self, other):
        if isinstance(other, Statistic):
            def a_mod_b(*x):
                return self(*x) % other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_mod_b(*x):
                return self(*x) % f(*x)
            label = str(other)
        elif self.codim == 1:
            def a_mod_b(*x):
                try:
                    return scalarize(self(*x)) % other
                except Exception as e:
                    raise OperationError(f'Could not compute {self.name} % {other}: {str(e)}')
            label = str(other)
        else:
            def a_mod_b(*x):
                val = self(*x)
                if len(val) != 1:
                    raise OperationError(f'Statistic {self.name} is not a scalar but % requires it; try using Scalar explicitly.')
                try:
                    return scalarize(self(*x)) % other
                except Exception as e:
                    raise OperationError(f'Could not compute {self.name} % {other}: {str(e)}')
            label = str(other)
        return Statistic(a_mod_b, dim=0, name=f'{self.name}(__) % {label}')

    def __rmod__(self, other):
        if callable(other):   # other cannot be a Statistic in __r*__
            f = tuple_safe(other)
            def a_mod_b(*x):
                return f(*x) % self(*x)
        else:
            def a_mod_b(*x):
                return other % self(*x)

        return Statistic(a_mod_b, dim=0, name=f'{str(other)} % {self.name}(__)')

    def __pow__(self, other):
        if isinstance(other, Statistic):
            def a_pow_b(*x):
                return self(*x) ** other(*x)
            label = f'{other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_pow_b(*x):
                return self(*x) ** f(*x)
            label = str(other)
        else:
            def a_pow_b(*x):
                return self(*x) ** as_numeric(other)
            label = str(other)

        return Statistic(a_pow_b, dim=0, name=f'{self.name}(__) ** {label}')

    def __rpow__(self, other):
        if callable(other):   # other cannot be a Statistic in __r*__
            f = tuple_safe(other)
            def a_pow_b(*x):
                return f(*x) ** self(*x)
        else:
            def a_pow_b(*x):
                return as_numeric(other) ** self(*x)

        return Statistic(a_pow_b, dim=0, name=f'{str(other)} ** {self.name}(__)')

    def __and__(self, other):
        if isinstance(other, Statistic):
            def a_and_b(*x):
                return self(*x) and other(*x)
            label = f'{self.name}(__) and {other.name}'
        elif callable(other):
            f = tuple_safe(other)
            def a_and_b(*x):
                return self(*x) and f(*x)
            label = f'{self.name}(__) and {str(other)}'
        else:
            def a_and_b(*x):
                return self(*x) and other
            label = f'{self.name}(__) and {str(other)}'

        return Statistic(a_and_b, dim=0, name=label)

    def __or__(self, other):
        if isinstance(other, Statistic):
            def a_or_b(*x):
                return self(*x) or other(*x)
            label = f'{self.name}(__) or {other.name}(__)'
        elif callable(other):
            f = tuple_safe(other)
            def a_or_b(*x):
                return self(*x) or f(*x)
            label = f'{self.name}(__) or {str(other)}'
        else:
            def a_or_b(*x):
                return self(*x) or other
            label = f'{self.name} and {str(other)}'

        return Statistic(a_or_b, dim=0, name=label)


def is_statistic(x) -> TypeGuard[Statistic]:
    return isinstance(x, Statistic)

class MonoidalStatistic(Statistic):
    def __init__(self,
                 fn: Callable | 'Statistic',        # Either a Statistic or a function to be turned into one
                 unit,                              # The unit of the monoid
                 dim: Optional[int] = None,       # Number of arguments the function takes; 0 means tuple expected
                 codim: Optional[int] = None,       # Dimension of the codomain; None means don't know
                 name: Optional[str] = None,        # A user-facing name for the statistic
                 description: Optional[str] = None  # A description used as a __doc__ string for the Statistic
                ) -> None:
        super().__init__(fn, dim, codim, name, description)
        self.unit = unit

    def __call__(self, *args):
        if len(args) == 0:
            return self.unit
        return super().__call__(*args)

class ProjectionStatistic(Statistic, Projection):
    def __init__(self,
                 # ATTN: Don't need this here, just adapt project
                 fn: Callable | 'Statistic',         # Either a Statistic or a function to be turned into one ATTN: No need for fn here!
                 onto: Iterable[int] | slice | Self, # 1-indexed projection indices
                 name: Optional[str] = None          # A user-facing name for the statistic
                ) -> None:
        codim = None
        dim = 0
        if isinstance(onto, ProjectionStatistic):
            indices: Iterable[int] | slice | 'ProjectionStatistic' = onto.subspace
            codim = onto.codim
            label = onto.name.replace('project[', '').replace(']', '')

        if isinstance(onto, Iterable):
            indices = list(onto)
            codim = len(indices)
            dim = max(indices)
            label = ", ".join(map(str, indices))
            if any([index == 0 for index in indices]):  # Negative from the end OK
                raise StatisticError('Projection indices are 1-indexed and must be non-zero')
        elif isinstance(onto, slice):
            indices = onto
            has_step = indices.step is None
            label = f'{indices.start or ""}:{indices.stop or ""}{":" if has_step else ""}{indices.step if has_step else ""}'
            # ATTN: Already converted in project; need to merge this
            #if indices.start == 0 or indices.stop == 0:
            #    raise StatisticError('Projection indices are 1-indexed and must be non-zero')
        
        description = textwrap.wrap(f'''A statistic that projects any value of dimension >= {dim or 1}
                                        to extract the {codim} components with indices {label}.''')
        # ATTN: Just pass project here, don't take an fn arg!
        super().__init__(fn, 0, codim, name, '\n'.join(description))
        self._components = indices

    @property
    def subspace(self):
        return self._components

    #ATTN: Make project() below a method here
    #ATTN?? Add minimum_dim property that specifies minimum compatible dimension; e.g., Project[3] -> 3, Project[2:-1] -> 2, Project[1,3,5] -> 5

def _ibool(x) -> Literal[0, 1]:
    return 1 if bool(x) else 0

class Condition(Statistic):
    """A condition is a statistic that returns a boolean value.

    Boolean values here are represented in the output with
    0 for false and 1 for true, though the input callable
    can return any
    """
    def __init__(
            self,
            predicate: Callable | 'Statistic',  # Either a Statistic or a function to be turned into one
            dim: Optional[int] = None,          # Number of arguments the function takes; 0 means tuple expected
            name: Optional[str] = None,         # A user-facing name for the statistic
            description: Optional[str] = None   # A description used as a __doc__ string for the Statistic
    ) -> None:
        super().__init__(predicate, dim, 1, name, description)
        self.__doc__ = self.__describe__(description or predicate.__doc__ or '', 'returns a 0-1 (boolean) value')

    def __call__(self, *args) -> Literal[0, 1] | Statistic:
        if len(args) == 1 and isinstance(args[0], Transformable):
            return args[0].transform(self)
        if isinstance(args[0], Statistic):
            return Condition(compose2(self, args[0]))
        result = super().__call__(*args)
        return _ibool(as_scalar(result))
        # if is_vec_tuple(result):
        #     return result.map(_ibool)
        # return as_vec_tuple(result).map(_ibool)


#
# Statistic decorator for easily creating a statistic out of a function
#

def statistic(
        maybe_fn: Optional[Callable] = None,  # If supplied, return Statistic, else a decorator
        *,
        name: Optional[str] = None,         # A user-facing name for the statistic
        dim: Optional[int] = None,          # Number of arguments the function takes; 0 means tuple expected
        codim: Optional[int] = None,        # Dimension of the codomain; None means don't know
        description: Optional[str] = None,  # A description used as a __doc__ string for the Statistic
        monoidal=None                       # If not None, the unit for a Monoidal Statistic
) -> Statistic | Callable[[Callable], Statistic]:
    """
    """
    if maybe_fn and monoidal is None:
        return Statistic(maybe_fn, dim, codim, name, description)
    elif maybe_fn:
        return MonoidalStatistic(maybe_fn, monoidal, dim, codim, name, description)

    if monoidal is None:
        def decorator(fn: Callable) -> Statistic:     # Function to be converted to a statistic
            return Statistic(fn, dim, codim, name, description)
    else:
        def decorator(fn: Callable) -> Statistic:     # Function to be converted to a statistic
            return MonoidalStatistic(fn, monoidal, dim, codim, name, description)
    return decorator

def scalar_statistic(
        name: Optional[str] = None,         # A user-facing name for the statistic
        dim: Optional[int] = None,          # Number of arguments the function takes; 0 means tuple expected
        description: Optional[str] = None,  # A description used as a __doc__ string for the Statistic
        monoidal=None                     # If not None, the unit of a Monoidal Statistic
):
    def decorator(fn: Callable) -> Statistic:     # Function to be converted to a statistic
        return Statistic(fn, dim, 1, name, description)
    return decorator

def condition(
        maybe_predicate: Optional[Callable] = None,  # If supplied, return Condition, else a decorator
        *,
        name: Optional[str] = None,         # A user-facing name for the statistic
        dim: Optional[int] = None,          # Number of arguments the function takes; 0 means tuple expected
        codim: Optional[int] = None,        # Dimension of the codomain; None means don't know
        description: Optional[str] = None,  # A description used as a __doc__ string for the Statistic
        monoidal=None                     # If not None, the unit for a Monoidal Statistic
) -> Condition | Callable[[Callable], Condition]:
    """
    """
    if maybe_predicate:
        return Condition(maybe_predicate, dim, name, description)

    def decorator(predicate: Callable) -> Condition:     # Function to be converted to a statistic
        return Condition(predicate, dim, name, description)
    return decorator


#
# Statistics Combinators
#

def fork(*statistics: Statistic) -> Statistic:
    arities = {stat.arity for stat in statistics}
    def forked(*x):
        return tuple([stat(*x) for stat in statistics])
    names = [stat.name for stat in statistics]
    return Statistic(forked, dim=next(iter(arities)) if len(arities) == 1 else 0,
                     name=f'fork({", ".join(names)})',
                     description=f'returns a tuple of the results of ({", ".join(names)})')

def chain(*statistics: Statistic) -> Statistic:
    "Compose statistics in pipeline order: (f ; g)(x) = g(f(x)), read 'f then g'."
    # ATTN: check arities compatible etc
    def chained(*x):
        state = x
        for stat in statistics:
            state = stat(*state)
        return state
    arity = statistics[0].arity if len(statistics) > 0 else None
    names = ", ".join([stat.name for stat in statistics])
    return Statistic(chained, arity, name=f'chain({names})')

def compose(*statistics: Statistic) -> Statistic:
    "Compose statistics in mathematical order: (f o g)(x) = f(g(x)), read 'f after g'."
    # ATTN: check dims and codims etc
    rev_statistics = list(statistics)
    rev_statistics.reverse()
    def composed(*x):
        state = x
        for stat in rev_statistics:
            state = stat(*state)
        return state
    arity = rev_statistics[0].arity if len(statistics) > 0 else None
    names = ", ".join([stat.name for stat in statistics])
    return Statistic(composed, arity, name=f'compose({names})')


#
# Special Numerical Values
#

infinity = math.inf  # ATTN: if needed, convert to appropriate value component type


#
# Commonly Used Statistics
#

Id = Statistic(identity, dim=0, name='identity', description='returns the value given as is')
Scalar = Statistic(lambda x: x[0] if is_tuple(x) else x, dim=1, name='scalar', description='represents a scalar value')
__ = Id # ATTN  Make this act like both Id and Scalar
_x_ = Scalar

Sum = MonoidalStatistic(sum, unit=0, dim=0, codim=1, name='sum', description='returns the sum of all the components of the given value')
Count = MonoidalStatistic(len,unit=0, dim=0, codim=1, name='count', description='returns the number of components in the given value')
Max = MonoidalStatistic(max, unit=-math.inf, dim=0, codim=1, name='max', description='returns the maximum of all components of the given value')
Min = MonoidalStatistic(min, unit=math.inf, dim=0, codim=1, name='min', description='returns the minimum of all components of the given value')
Mean = Statistic(lambda x: sum(x)/len(x), dim=0, codim=1, name='mean', description='returns the arithmetic mean of all components of the given value')
Abs = Statistic(abs, dim=1, codim=1, name='abs', description='returns the absolute value of the given number')

# Diff == takes first order diffs of the tuple
# Diffs(k) takes k-order diffs

def Constantly(x) -> Statistic:
    xtuple = ensure_tuple(x)
    return Statistic(lambda _: xtuple, dim=0, codim=len(xtuple), name=f'The constant {xtuple}')

Cos = Statistic(math.cos, dim=1, codim=1, name='cos', description='returns the cosine of a scalar argument')
Sin = Statistic(math.sin, dim=1, codim=1, name='sin', description='returns the sine of a scalar argument')
Exp = Statistic(math.exp, dim=1, codim=1, name='exp', description='returns the exponential of a scalar argument')
Log = Statistic(math.log, dim=1, codim=1, name='log', description='returns the natural logarithm of a positive scalar argument')
Log2 = Statistic(math.log2, dim=1, codim=1, name='log', description='returns the logarithm base 2 of a positive scalar argument')
Log10 = Statistic(math.log2, dim=1, codim=1, name='log', description='returns the logarithm base 10 of a positive scalar argument')


#
# Combinators
#

def ForEach(s: Statistic) -> Statistic:
    def foreach(*x):
        if len(x) > 0 and is_tuple(x[0]):
            x = x[0]
        return tuple([s(xi) for xi in x])
    return Statistic(foreach, dim=0, name=f'applies {s.name} to every component of input value')


# #
# # ATTN: Need better handling of dim, arity, codim.  arity=0, etc. is fine while dim is the minimum dimension allowed
# # codim=0 means returns a tuple whose equals its input length; codim=n means returns an n-tuple, codim=None is unknown
# #
def Fork(stat: Statistic, *more_stats: Statistic) -> Statistic:
    # Arities must all be the same
    if stat.arity != 0 and any([s.arity != 0 and s.arity != stat.arity for s in more_stats]):  # ATTN!: Statistics need to distinguish arity from minimum dim, need two properties
        raise ValueError('Fork must be called on statistics with the same dimension')
    codim = 0
    if stat.codim is not None and stat.codim > 0 and all([s.codim is not None and s.codim > 0 for s in more_stats]):
        codim = stat.codim + sum(s.codim for s in more_stats)  # type: ignore
    def forked(*x):
        returns = list(stat(x))
        for s in more_stats:
            returns.extend(s(x))
        return returns
    return Statistic(forked, dim=stat.arity, codim=codim, name=f'fork({stat.name}, {", ".join([s.name for s in more_stats])})')

## ATTN: Add 
# Permute    ATTN: fix up but keeping it simple for now
def Permute(p: Iterable[int]):
    # assert p contains all unique values from 1..n
    perm = list(map(lambda k: k - 1, p))
    n = len(perm)
    @statistic(name=f'Permutation', dim=0)   # ATTN
    def permute(value):
        return VecTuple(value[perm[i]] for i in range(n))
    return permute
        


def IfThenElse( cond: Statistic
              , t: Statistic | tuple | float | int
              , f: Statistic | tuple | float | int
              ) -> Statistic:
    if not is_statistic(t):
        t = Constantly(t)
    if not is_statistic(f):
        f = Constantly(f)
    # ATTN: arity conditions need to be more nuanced
    #if cond.arity != t.arity or t.arity != f.arity or t.codim != f.codim:
    if t.codim is not None and f.codim is not None and t.codim != f.codim:
        raise ValueError(f'True and False statistics for IfElse must have matching dim and codim')  # ATTN:Replace with custom error for catching at repl
    def ifelse(*x):
        if cond(*x):
            return t(*x)
        else:
            return f(*x)
    return Statistic(ifelse, dim=cond.arity, codim=t.codim, name=f'returns {t.name} if {cond.name} is true else returns {f.name}')
        
def Not(s: Statistic) -> Condition:
    #ATTN: require s.codim == 1
    return Condition(lambda *x: 1 - s(*x), dim=s.arity, name=f'not({s.name})', description=f'returns the logical not of {s.name}')

# def And(s1: Statistic, s2: Statistic) -> Statistic:
#     #ATTN: require si.codim == 1 and s1.arity == s2.arity
#     return Statistic(lambda *x: s1(*x) and s2(*x), dim=s1.arity, codim=1, name=f'Logical And of {s1.name} and {s2.name}')
#  
# def Or(s1: Statistic, s2: Statistic) -> Statistic:
#     #ATTN: require si.codim == 1 and s1.arity == s2.arity
#     return Statistic(lambda *x: s1(*x) or s2(*x), dim=s1.arity, codim=1, name=f'Logical Or of {s1.name} and {s2.name}')

def And(*stats: Statistic) -> Condition:
    #ATTN: require si.codim == 1 and s1.arity == s2.arity
    def and_of(*x):
        val = True
        for s in stats:
            val = val and s(*x)
            if not val:
                break
        return val
    labels = ["'" + s.name + "'" for s in stats]
    return Condition(and_of, dim=max(s.arity for s in stats), name=f'({" and ".join(labels)})', description=f'returns the logical and of {", ".join(labels)}')

def Or(*stats: Statistic) -> Condition:
    #ATTN: require si.codim == 1 and s1.arity == s2.arity
    def and_of(*x):
        val = False
        for s in stats:
            val = val or s(*x)
            if val:
                break
        return val
    labels = ["'" + s.name + "'" for s in stats]
    return Condition(and_of, dim=max(s.arity for s in stats), name=f'({" or ".join(labels)})', description=f'returns the logical or of {", ".join(labels)}')

top = Condition(lambda _x: True, name='top', description='returns true for any value')

bottom = Condition(lambda _x: False, name='bottom', description='returns false for any value')


## ATTN: These should really be methods of ProjectionStatistic
## There should be no need for a callable argment in that constructor.
@overload
def project(*__indices: int) -> ProjectionStatistic:
    ...

@overload
def project(__index_tuple: Iterable[int]) -> ProjectionStatistic:
    ...

def project(*indices_or_tuple) -> ProjectionStatistic:
    """Creates a projection statistic that extracts the specified components.

       Positional variadic arguments:
         *indices_or_tuple -- a tuple of integer indices starting from 1 or a single int tuple
    """
    if len(indices_or_tuple) == 0:  # ATTN:Error here instead?
        return ProjectionStatistic(lambda _: (), (), name='Null projection')

    # ATTN:Support slice objects here
    # In that sense, it would be good if the projection statistic could also get
    # the dimension of the input tuple, then we could use Proj[2:-1] to mean
    # all but the first and Proj[1:-2] for all but the last regardless of
    # dimension.

    if isinstance(indices_or_tuple[0], slice):
        def dec_or_none(x: int | None) -> int | None:
            if x is not None and x > 0:
                return x - 1
            return x
        zindexed = indices_or_tuple[0]
        indices: slice | Iterable = slice(dec_or_none(zindexed.start),
                                          dec_or_none(zindexed.stop),
                                          zindexed.step)
        def get_indices(xs):
            return as_vec_tuple(xs[indices])
        label = str(indices)
    else:
        if isinstance(indices_or_tuple[0], Iterable):
            indices = indices_or_tuple[0]
        else:
            indices = indices_or_tuple

        def get_indices(xs):
            getter = itemgetter(*[x - 1 if x > 0 else x for x in indices if x != 0])
            return as_vec_tuple(getter(xs))
        label = ", ".join(map(str, indices))
    return ProjectionStatistic(
        get_indices,
        indices,
        name=f'project[{label}]')


class ProjectionFactory:
    @overload
    def __call__(self, *__indices: int) -> ProjectionStatistic:
        ...
     
    @overload
    def __call__(self, __index_tuple: Iterable[int]) -> ProjectionStatistic:
        ...
     
    @overload
    def __call__(self, __index_slice: slice) -> ProjectionStatistic:
        ...
     
    def __call__(self, *indices_or_tuple) -> ProjectionStatistic:
        return project(*indices_or_tuple)

    @overload
    def __getitem__(self, *__indices: int) -> ProjectionStatistic:
        ...
     
    @overload
    def __getitem__(self, __index_tuple: Iterable[int]) -> ProjectionStatistic:
        ...

    @overload
    def __getitem__(self, __index_slice: slice) -> ProjectionStatistic:
        ...
     
    def __getitem__(self, *indices_or_tuple) -> ProjectionStatistic:
        return project(*indices_or_tuple)

Proj = ProjectionFactory()




from __future__ import annotations

import re

from abc               import ABC
from collections       import defaultdict
from decimal           import Decimal
from operator          import add
from typing            import Callable, cast, Literal, Generator, Union  # ATTN: collections.abc.Generator ok for 3.9?
from typing_extensions import TypeGuard

from frplib.exceptions import ConstructionError
from frplib.numeric    import (Numeric, ScalarQ,
                               as_real, as_nice_numeric, as_numeric,
                               is_scalar_q, show_numeric)
# from frplib.protocols  import NamedCallable
# from frplib.utils      import every

#
# Helpers
#

def merge_with(a, b, merge_fn=lambda x, y: y):
    merged = {k: a.get(k, b.get(k)) for k in a.keys() ^ b.keys()}
    merged.update({k: merge_fn(a[k], b[k]) for k in a.keys() & b.keys()})
    return merged

def is_zero(x: Numeric):  # Move to Numeric, should it take ScalarQ?
    return x == 0  # Temp

def show_coef(x: ScalarQ) -> str:
    return show_numeric(as_numeric(x), max_denom=1)


#
# Unique Symbols
#

def symbol_name_generator(base='#x_') -> Generator[str, None, None]:
    script = 0
    while True:
        yield f'{base}{{{script}}}'
        script += 1

gensym = symbol_name_generator()


#
# Generic Symbolic Quantities
#

class Symbolic(ABC):
    def is_pure(self):
        ...

    def pure_value(self):
        ...

    @property
    def key_of(self):
        ...

    @property
    def sort_key(self):
        ...

    def substitute(self, mapping: dict[str, ScalarQ], purify=True):
        ...

    def __lt__(self, other):
        "Sort symbols before numbers and by sort_key against Symbolics."
        if isinstance(other, Symbolic):
            return self.sort_key < other.sort_key
        if isinstance(other, str):
            return str(self) < other
        if is_scalar_q(other):
            pv = self.pure_value()
            if pv is not None:
                return pv < as_real(other)
        return True


#
# Multinomial Terms
#

class SymbolicMulti(Symbolic):
    "A symbolic multinomial term c a_1^k_1 a_2^k_2 ... a_n^k_n."
    def __init__(self, vars: list[str], powers: list[int], coef: ScalarQ = 1):
        # if powers too short, those count as 0, so ok; extra powers ignored
        # coef, [], []  acts as a scalar and a *multiplicative* identity
        self.coef = as_numeric(coef)

        order = 0
        sig = show_coef(self.coef)
        multi: dict[str, int] = defaultdict(int)
        if not is_zero(self.coef):
            for var, pow in zip(vars, powers):
                if not var:
                    raise ConstructionError('A symbolic variable name must be a non-empty string.')
                multi[var] += pow
                order += pow
            sigs = []
            for var, pow in sorted(multi.items()):
                if pow == 0:
                    del multi[var]
                    continue
                sigs.append(f'{var}^{pow}')
            if sigs:
                sig = " ".join(sigs)

        self.term = multi
        self.order = order
        self.key = sig
        self.as_str: Union[str, None] = None   # Computed lazily

    @classmethod
    def from_terms(cls, multi: dict[str, int], coef: Numeric = 1):
        if is_zero(coef):
            return SymbolicMulti([], [], 0)
        vars = []
        pows = []
        for var, pow in multi.items():
            if pow != 0:
                vars.append(var)
                pows.append(pow)
        return cls(vars, pows, coef)

    @classmethod
    def pure(cls, x: ScalarQ = 0):
        return cls([], [], as_numeric(x))

    def is_pure(self) -> bool:
        return len(self.term) == 0

    def pure_value(self):
        if self.is_pure():
            return self.coef
        return None

    def substitute(self, mapping: dict[str, ScalarQ], purify=True) -> SymbolicMulti | Numeric:
        coef = self.coef
        multi = self.term.copy()
        for var, pow in self.term.items():
            if var in mapping:
                coef *= as_numeric(mapping[var]) ** pow
                del multi[var]
        coef = as_numeric(coef)  # ATTN: needed?
        if len(multi) == 0:
            return coef
        return SymbolicMulti.from_terms(multi, coef)

    @property
    def signature(self) -> str:
        return self.key

    @property
    def key_of(self):
        return self.key

    @property
    def sort_key(self):
        return (self.coef, self.key)

    def __bool__(self) -> bool:
        return False if self.is_pure() and is_zero(self.coef) else True

    def __str__(self) -> str:
        if self.as_str is None:
            if self.is_pure():
                self.as_str = show_coef(self.coef)
            else:
                coef = '' if self.coef == 1 else show_coef(self.coef) + ' '
                term = re.sub(r'\^1( |$)', lambda m: ' ' if m.group(1) else '', self.signature)
                self.as_str = coef + term
        return self.as_str

    def __frplib_repr__(self) -> str:
        return str(self)

    def clone(self) -> SymbolicMulti:
        return self.from_terms(self.term, self.coef)

    def __mul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                return 0
            return self.from_terms(self.term, self.coef * as_real(other))
        if isinstance(other, SymbolicMulti):
            term = other.term if self.is_pure() else merge_with(self.term, other.term, add)
            coef = self.coef * other.coef
            return SymbolicMulti.from_terms(term, coef)
        return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                return 0
            return self.from_terms(self.term, as_real(other) * self.coef)
        # Cannot be SymbolicMulti in rul
        return NotImplemented

    def __add__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if self.is_pure():
                return self.from_terms(self.term, self.coef + as_real(other))
            return SymbolicMultiSum([self, SymbolicMulti.pure(as_numeric(other))])
        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([self, other])
        return NotImplemented

    def __radd__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if self.is_pure():
                return self.from_terms(self.term, as_real(other) + self.coef)
            return SymbolicMultiSum([SymbolicMulti.pure(as_numeric(other)), self])
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if self.is_pure():
                return self.from_terms(self.term, self.coef - as_real(other))
            return SymbolicMultiSum([self, SymbolicMulti.pure(as_numeric(-other))])
        if isinstance(other, SymbolicMulti):
            if self.key_of == other.key_of:
                return as_numeric(self.coef - other.coef)
            return SymbolicMultiSum([self, -1 * other])
        return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if self.is_pure():
                return self.from_terms(self.term, as_real(other) - self.coef)
            return SymbolicMultiSum([SymbolicMulti.pure(as_numeric(other)), -1 * self])
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            return self.from_terms(self.term, self.coef / as_real(other))

        if isinstance(other, (SymbolicMulti, SymbolicMultiSum)):
            return symbolic(self, other)

        if isinstance(other, SymbolicMultiRatio):
            return symbolic(self * other.denominator, other.numerator)

        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            return symbolic(SymbolicMulti.pure(other), self)

        if isinstance(other, (SymbolicMulti, SymbolicMultiSum)):
            return symbolic(other, self)

        if isinstance(other, SymbolicMultiRatio):
            return symbolic(other.numerator, self * other.denominator)

        return NotImplemented

    def __pow__(self, n):
        if isinstance(n, int):
            if n == 0:
                return 1
            term = self.term.copy()
            for var, pow in term.items():
                term[var] = pow * n if var else pow   # Constant always has power 1
            return self.from_terms(term, self.coef ** n)
        return NotImplemented

symbolic_zero = SymbolicMulti.pure(0)
symbolic_one = SymbolicMulti.pure(1)


#
# Symbolic Function Calls
#

# class SymbolicFn(SymbolicMulti):
#     def __init__(self, fn: NamedCallable, args: list[Symbolic]) -> None:
#         var = next(gensym)
#         self.fns = {var: {'fn': fn, 'args': args[:]}}
#         super().__init__([var], [1])
#
#     def is_pure(self) -> bool:
#         return (super().is_pure() or
#                 (self.term.keys() == self.fns.keys() and
#                  all(every(lambda s: s.is_pure(), f['args']) for f in self.fns.values())))
#
#     def pure_value(self):
#         if self.is_pure():
#             value = self.coef
#             for fn_key in self.fns:
#                 fstar = self.fns[fn_key]
#                 fn = cast(Callable[..., Numeric], fstar['fn'])
#                 args: list = fstar['args']
#                 assert callable(fn)
#                 value *= as_real(fn(*args))
#             return value
#         return None
#
#     def substitute(self, mapping: dict[str, ScalarQ], purify=True) -> SymbolicFn | Numeric:
#         multi = self.substitute(mapping, purify)
#         pass
#
#     def __mul__(self, other):
#         # Stay as SymbolicFn against SymbolicMulti and pure
#         pass
#
#     def __rmul__(self, other):
#         # Stay as SymbolicFn against SymbolicMulti and pure
#         pass


#
# Sums of Multinomial Terms
#

class SymbolicMultiSum(Symbolic):
    "A sum of symbolic multinomial terms sum_i c_i a_i1^k_1 a_i2^k_2 ... a_in^k_n."
    def __init__(self, multis: list[SymbolicMulti]) -> None:
        terms = self.combine_terms(multis)
        if not terms:
            self.terms: list[SymbolicMulti] = []
            self.coef: Numeric = 0
            self.order = 0
            self.key = '0'
            self.as_str: str | None = '0'
        elif len(terms) == 1 and terms[0].is_pure():
            self.terms = []
            self.coef = terms[0].pure_value()
            self.order = 0
            self.key = '1'
            self.as_str = show_coef(self.coef)
        else:
            terms = sorted(terms, key=lambda t: t.signature)
            self.order = 0
            for term in terms:
                if term.order > self.order:
                    self.order = term.order
                    self.coef = term.coef
            self.key = " + ".join([str(term / self.coef) for term in terms])
            self.terms = terms
            self.as_str = None

    def __str__(self) -> str:
        if self.as_str is None:
            self.as_str = " + ".join([str(term) for term in self.terms])
        return self.as_str

    def __frplib_repr__(self) -> str:
        return str(self)

    def is_pure(self) -> bool:
        return len(self.terms) == 0

    def is_single(self) -> bool:
        return len(self.terms) == 1

    def pure_value(self):
        if self.is_pure():
            return self.coef
        return None

    def substitute(self, mapping: dict[str, ScalarQ], purify=True) -> SymbolicMultiSum | Numeric:
        if self.is_pure() and purify:
            return as_nice_numeric(self.coef)
        total: Union[SymbolicMultiSum, Numeric] = sum(term.substitute(mapping, purify=purify)  # type: ignore
                                                      for term in self.terms)
        if isinstance(total, Symbolic):
            return total
        return as_nice_numeric(total)

    @property
    def key_of(self):
        return self.key

    @property
    def sort_key(self):
        return (self.coef, self.key)

    @classmethod
    def singleton(cls, sym: SymbolicMulti):
        return SymbolicMultiSum([sym])

    @staticmethod
    def combine_terms(terms: list[SymbolicMulti]) -> list[SymbolicMulti]:
        combined: dict[str, SymbolicMulti] = {}
        for term in terms:
            k = term.signature
            if k in combined:
                combined[k] = SymbolicMulti.from_terms(term.term, combined[k].coef + term.coef)
            else:
                combined[k] = term
        return [v for v in combined.values() if not is_zero(v.coef)]

    def __add__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            return SymbolicMultiSum([*self.terms, SymbolicMulti.pure(other)])

        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([*self.terms, other])

        if isinstance(other, SymbolicMultiSum):
            return SymbolicMultiSum([*self.terms, *other.terms])
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            return SymbolicMultiSum([*self.terms, SymbolicMulti.pure(-other)])

        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([*self.terms, -1 * other])

        if isinstance(other, SymbolicMultiSum):
            terms = [*self.terms]
            terms.extend(-1 * term for term in other.terms)
            return SymbolicMultiSum(terms)
        return NotImplemented

    def __radd__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            return SymbolicMultiSum([SymbolicMulti.pure(other), *self.terms])

        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([other, *self.terms])

        return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            terms = [SymbolicMulti.pure(-other)]
            terms.extend(-1 * term for term in self.terms)
            return SymbolicMultiSum(terms)

        if isinstance(other, SymbolicMulti):
            terms = [other]
            terms.extend(-1 * term for term in self.terms)
            return SymbolicMultiSum(terms)

        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            nother = as_numeric(other)
            if is_zero(nother):
                return 0
            if nother == 1:
                return self
            pv = self.pure_value()
            if pv is not None:
                return pv * nother
            return SymbolicMultiSum([x * as_real(nother) for x in self.terms])

        pv = self.pure_value()
        if pv is not None:
            return pv * other

        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([x * other for x in self.terms])

        if isinstance(other, SymbolicMultiSum):
            ov = other.pure_value()
            if ov is not None:
                return self * ov

            # Combine like terms
            combined: dict[str, SymbolicMulti] = {}
            for term1 in self.terms:
                for term2 in other.terms:
                    prod = term1 * term2
                    k = prod.signature
                    if k in combined:
                        combined[k].coef += prod.coef
                    else:
                        combined[k] = prod
            terms = [v for v in combined.values() if not is_zero(v.coef)]
            return SymbolicMultiSum(terms)
        return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            nother = as_numeric(other)
            if is_zero(nother):
                return 0
            if nother == 1:
                return self
            pv = self.pure_value()
            if pv is not None:
                return nother * pv
            return SymbolicMultiSum([as_real(nother) * x for x in self.terms])

        pv = self.pure_value()
        if pv is not None:
            return other * pv

        if isinstance(other, SymbolicMulti):
            return SymbolicMultiSum([other * x for x in self.terms])

        return NotImplemented

    def __pow__(self, n):
        if isinstance(n, int):
            if n == 0:
                return 1
            if n == 1:
                return self
            if n % 2 == 0:
                return (self * self) ** (n // 2)
            return self * (self * self) ** (n // 2)
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if other == 1:
                return self
            d = as_real(other)
            return SymbolicMultiSum([term / d for term in self.terms])

        if isinstance(other, (SymbolicMulti, SymbolicMultiSum, SymbolicMultiRatio)):
            opv = other.pure_value()
            if opv is not None:
                if opv == 1:
                    return self
                return SymbolicMultiSum([term / opv for term in self.terms])
            return simplify(symbolic(self, other))

        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if as_numeric(other) == 1:
                return symbolic(symbolic_one, self)
            d = as_real(other)
            return symbolic(SymbolicMulti.pure(d), self)

        if isinstance(other, (SymbolicMulti, SymbolicMultiRatio)):
            return simplify(symbolic(other, self))

        return NotImplemented


#
# Ratios of Sums of Multinomial Terms
#

class SymbolicMultiRatio(Symbolic):
    def __init__(self, numerator: SymbolicMultiSum, denominator: SymbolicMultiSum):
        terms = [numerator, denominator]
        # Simplify in the generic symbolic constructor so we can return
        # other types of values.
        self.terms = terms
        self.key = f'({numerator.key_of})/({denominator.key_of})'

        if numerator.is_pure() or numerator.is_single():
            num = str(numerator)
        else:
            num = f'({str(numerator)})'

        if denominator.is_pure() or denominator.is_single():
            den = str(denominator)
            if ' ' in den:
                den = f'({str(denominator)})'
        else:
            den = f'({str(denominator)})'

        self.as_str = f'{num}/{den}'

    @property
    def numerator(self):
        return self.terms[0]

    @property
    def denominator(self):
        return self.terms[1]

    @property
    def key_of(self):
        return self.key

    @property
    def sort_key(self):
        return (as_real(self.numerator.coef) / as_real(self.denominator.coef), self.key)

    def __str__(self) -> str:
        return self.as_str

    def __frplib_repr__(self) -> str:
        return str(self)

    def is_pure(self) -> bool:
        return self.numerator.is_pure() and self.denominator.is_pure()

    def pure_value(self):
        npv = self.numerator.pure_value()
        dpv = self.denominator.pure_value()

        if npv is not None and dpv is not None:
            return as_numeric(npv / dpv)
        return None

    def substitute(self, mapping: dict[str, ScalarQ], purify=True) -> Symbolic | Numeric:
        spv = self.pure_value()
        if spv is not None:
            return spv if purify else SymbolicMultiSum.singleton(SymbolicMulti.pure(spv))

        num = self.numerator.substitute(mapping, purify)
        den = self.denominator.substitute(mapping, purify)

        if not isinstance(num, SymbolicMultiSum) and not isinstance(num, SymbolicMultiSum):
            return as_nice_numeric(as_real(num / den))
        return simplify(symbolic(num, den))

    def __add__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                simplify(self)
            numer = self.numerator + as_real(other) * self.denominator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiRatio):
            if self.denominator.key_of == other.denominator.key_of:
                return simplify(SymbolicMultiRatio(self.numerator + other.numerator, self.denominator))

            numer = self.numerator * other.denominator + self.denominator * other.numerator
            denom = self.denominator * other.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiSum):
            numer = self.numerator + other * self.denominator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            if not other:
                return self
            numer = self.numerator + SymbolicMultiSum.singleton(other) * self.denominator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __radd__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                self
            numer = as_real(other) * self.denominator + self.numerator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiSum):
            numer = other * self.denominator + self.numerator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            if not other:
                return self
            numer = SymbolicMultiSum.singleton(other) * self.denominator + self.numerator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __sub__(self, other):
        return self + (-1 * other)

    def __rsub__(self, other):
        return other + (-1 * self)

    def __mul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                return symbolic_zero
            numer = self.numerator * as_real(other)
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiRatio):
            if self.numerator.key_of == other.denominator.key_of:
                return simplify(symbolic(other.numerator, self.denominator))

            if self.denominator.key_of == other.numerator.key_of:
                return simplify(symbolic(self.numerator, other.denominator))

            numer = self.numerator * other.numerator
            denom = self.denominator * other.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiSum):
            numer = self.numerator * other
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            numer = self.numerator * SymbolicMultiSum.singleton(other)
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if is_zero(as_numeric(other)):
                return symbolic_zero
            numer = as_real(other) * self.numerator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiSum):
            numer = other * self.numerator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            numer = SymbolicMultiSum.singleton(other) * self.numerator
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            if other == 1:
                return self
            numer = self.numerator / as_numeric(other)
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiRatio):
            numer = self.numerator * other.denominator
            denom = self.denominator * other.numerator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMultiSum):
            numer = self.numerator
            denom = self.denominator * other
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            numer = self.numerator
            denom = self.denominator * SymbolicMultiSum.singleton(other)
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, (int, float, Decimal)):   # is_scalar_q(other):
            numer = SymbolicMultiSum.singleton(SymbolicMulti.pure(as_numeric(other)))
            denom = self.denominator
            return simplify(SymbolicMultiRatio(numer * denom, self.numerator))

        if isinstance(other, SymbolicMultiSum):
            numer = self.denominator * other
            denom = self.numerator
            return simplify(SymbolicMultiRatio(numer, denom))

        if isinstance(other, SymbolicMulti):
            numer = self.denominator * SymbolicMultiSum.singleton(other)
            denom = self.numerator
            return simplify(SymbolicMultiRatio(numer, denom))

        return NotImplemented

    def __pow__(self, n):
        if isinstance(n, int):
            if n == 0:
                return 1
            if n == 1:
                return self
            return simplify(symbolic(self.numerator ** n, self.denominator ** n))
        return NotImplemented


#
# Simple Simplification Rules
#

def simplify(a: Symbolic | Numeric) -> Union[Symbolic, Numeric]:
    if isinstance(a, (int, Decimal)):
        return a

    apv = a.pure_value()
    if apv is not None:
        return as_numeric(apv)

    if isinstance(a, (SymbolicMulti, SymbolicMultiSum)):
        return a

    assert isinstance(a, SymbolicMultiRatio)
    npv = a.numerator.pure_value()
    dpv = a.denominator.pure_value()

    if npv is not None:
        if is_zero(npv):
            return 0
        return SymbolicMultiRatio(SymbolicMultiSum.singleton(symbolic_one), a.denominator / as_real(npv))
    if dpv is not None:
        return a.numerator / as_real(dpv)

    if a.numerator.key_of == a.denominator.key_of:
        return as_real(a.numerator.coef / a.denominator.coef)

    # Other simplification rules?
    return a


#
# Symbolic Constructors (use these not the class constructors)
#

def symbolic(numerator: Union[Symbolic, str], denominator: Union[Symbolic, Literal[1]] = 1) -> Union[Symbolic, Numeric]:
    if isinstance(numerator, str):
        numerator = symbol(numerator)

    npv = numerator.pure_value()
    dpv = denominator.pure_value() if denominator != 1 else 1
    if npv is not None and dpv is not None:
        return as_real(npv / dpv)
    elif npv is not None and denominator == 1:
        return npv
    elif npv is not None:
        numerator = symbolic_one
        denominator = denominator / npv
    elif denominator == 1:
        return simplify(numerator)

    assert isinstance(numerator, Symbolic)
    assert isinstance(denominator, Symbolic)

    if isinstance(numerator, SymbolicMulti):
        numerator = SymbolicMultiSum.singleton(numerator)

    if isinstance(denominator, SymbolicMulti):
        denominator = SymbolicMultiSum.singleton(denominator)

    num_rat = isinstance(numerator, SymbolicMultiRatio)
    den_rat = isinstance(denominator, SymbolicMultiRatio)

    if num_rat and den_rat and numerator.key_of == denominator.key_of:        # type: ignore
        return ((numerator.numerator.coef * denominator.denominator.coef) /   # type: ignore
                (numerator.denominator.coef * denominator.numerator.coef))    # type: ignore
    if num_rat:
        numerator = cast(SymbolicMultiSum, numerator)
        return simplify(numerator.__truediv__(denominator))
    elif den_rat:
        denominator = cast(SymbolicMultiSum, denominator)
        return simplify(denominator.__rtruediv__(numerator))

    numerator = cast(SymbolicMultiSum, numerator)
    denominator = cast(SymbolicMultiSum, denominator)

    if numerator.key_of == denominator.key_of:
        return as_real(numerator.coef / denominator.coef)

    return simplify(SymbolicMultiRatio(numerator, denominator))
#
# Symbol constructors
#

def gen_symbol() -> SymbolicMulti:
    "Generates a symbol with a unique name."
    var = next(gensym)
    return SymbolicMulti([var], [1])

def symbol(var: str) -> SymbolicMulti:
    "Generates a symbol with the given name, typically a single letter."
    return SymbolicMulti([var], [1])

def is_symbolic(obj) -> TypeGuard[SymbolicMulti | SymbolicMultiSum | SymbolicMultiRatio]:
    return isinstance(obj, (SymbolicMulti, SymbolicMultiSum, SymbolicMultiRatio))



from pygments.lexer import RegexLexer
from pygments.token import (
    Punctuation,
    Operator,
    Other,
    Token,
    Whitespace,
)

from frplib.numeric import numeric_re, integer_re

class MarketCommandLexer(RegexLexer):
    name = 'MarketCommand'
    tokens = {
        'root': [
            (r'\b(?:demo|buy|show|compare)', Token.Command),
            (r'(?:with)', Token.Connective),
            (r'(?:kind)', Token.Kind),
            (r'[@.]', Operator),
            (r'[()]', Punctuation, 'kind'),
            (integer_re, Token.Count),
            (r'\s+', Whitespace),
            (r'.', Other),
        ],
        'kind': [
            (fr'<(?:{numeric_re}(?:\s*,\s*{numeric_re})*)?>', Token.Node),
            (r'[(]', Punctuation, '#push'),
            (r'[)]', Punctuation, '#pop'),
            (numeric_re, Token.Weight),
        ],
    }



#
# Monty Hall Example as described in the text
#

from frplib.calculate  import substitute_with
from frplib.kinds      import uniform, arbitrary
from frplib.statistics import statistic, Not
from frplib.utils      import identity


# The basic choices in the game, first Monty's then yours.

door_with_prize = uniform(1, 2, 3)
chosen_door = arbitrary(1, 2, 3, names=['l', 'm', 'r'])  # l (ell) might look funny in output


# Statitics used to decide whether you win under a strategy

@statistic
def got_prize_door_initially(x):
    monty, you = x
    return monty == you

didnt_get_prize_door_initially = Not(got_prize_door_initially)    # type: ignore


# The game outcome kind up to but not including the choice of whether to switch

game_outcome = door_with_prize * chosen_door


# The FRP that gives 1 if you win and 0 if you lose is an event.
# The kind of that event is determined by whether you switch
# or don't switch. The following are those kinds.

dont_switch_win = got_prize_door_initially(game_outcome)
switch_win = didnt_get_prize_door_initially(game_outcome)


# Analyze the game for any choice of doors
# The Win kinds are the same for any choice of left/middle/right weights.

def by_strategy(left=1, middle=1, right=1, *, switch=None):
    """Returns outcome for a specific strategy of yours.

    left, middle, right are weights you assign the three doors
      These should be positive numbers

    switch is None (the default), True, or False.
      If switch None, show the game outcome kind for your strategy up
      to but not including your decision to switch. If it is True,
      show the outcome where choose Switch, and if False, the outcome
      where you choose Don't Switch.

    """
    outcome = game_outcome.bimap(identity, substitute_with({'l': left, 'm': right, 'r': middle}))

    if switch is True:
        return got_prize_door_initially(outcome)

    if switch is False:
        return didnt_get_prize_door_initially(outcome)

    return outcome



#
# A singleton environment for capturing global options in an interactive session.
#
# This controls the output formats of objects that should print in nice
# or rich ways in an interactive session. While this can be used at the
# library level, it is primarily for interactive use and not thread safe.
#
from __future__ import annotations

from dataclasses  import dataclass

from rich.console import Console
from rich.theme   import Theme

bright_theme = Theme({
    "repr.number": "#3333cc",
    "repr.number_complex": "#333366",
    "repr.bool_true": "#009933",
    "repr.bool_false": "#990033",
    "repr.str": "#330066",
    "repr.attrib_name": "#330000",
    "repr.attrib_value": "#000033",
})

dark_theme = Theme({
    "repr.number": "#cccc33",
    "repr.number_complex": "#cccc99",
    "repr.bool_true": "#ff66cc",
    "repr.bool_false": "#66ffcc",
    "repr.str": "#ccff99",
    "repr.attrib_name": "#ccffff",
    "repr.attrib_value": "#ffffcc",
})


@dataclass
class Environment:
    """Options governing interactive sessions, globally available.
    """
    ascii_only: bool = False
    dark_mode: bool = False
    is_interactive: bool = False
    console: Console = Console(highlight=True, theme=bright_theme)

    def on_ascii_only(self) -> None:
        "Require ASCII-only output, no rich text, unicode, or markdown."
        self.ascii_only = True

    def off_ascii_only(self) -> None:
        "Allow non-ascii and rich output"
        self.ascii_only = False

    def on_dark_mode(self) -> None:
        "Changes text color to suit dark colored terminals"
        self.dark_mode = True
        self.console.push_theme(dark_theme)

    def on_bright_mode(self) -> None:
        "Text color default suited for light colored terminals"
        self.dark_mode = False
        self.console.push_theme(bright_theme)

    def interactive_mode(self, ascii=None) -> None:
        "Indicate that this session is interactive. No need to turn this off."
        self.is_interactive = True
        if ascii is not None:
            self.ascii_only = ascii

    def console_str(self, rich_str) -> str:
        with self.console.capture() as capture:
            self.console.print(rich_str)
        return capture.get()

environment = Environment()



from __future__ import annotations

from frplib.exceptions import IndexingError
from frplib.kinds      import Kind
from frplib.frps       import FRP
from frplib.statistics import Proj


#
# Utilities
#

def components(frp_or_kind: Kind | FRP):
    """Returns the components of an FRP or Kind.

    If X_ = components(X) for an FRP/Kind X, then X_[i] is the ith
    component FRP/Kind. Here, i must be between 1 and dim(X).
    This can also be used as an iterator:

        for x in components(X):
            ...

    In most cases, it is easier to subscript the FRP or Kind
    directly, like X[i].

    """
    class Component:
        def __init__(self, frp_or_kind: Kind | FRP, dim: int):
            self.this = frp_or_kind
            self.dim = dim

        def __getitem__(self, index):
            if index < 1 or index > self.dim:
                raise IndexingError(
                    f'The {index} component of this FRP/Kind does not exist; its dimension is {self.dim}.'
                )
            return Proj[self.dim](self.this)

        def __iter__(self):
            for i in range(1, self.dim + 1):
                yield self[i]

    return Component(frp_or_kind, frp_or_kind.dim)



from __future__        import annotations

from collections       import defaultdict
from functools         import reduce
from typing_extensions import Any

from parsy import (
    ParseError,
    regex,
    seq,
    string,
    whitespace,
)

from frplib.exceptions           import KindError
from frplib.numeric              import as_numeric, integer_re, numeric_q_from_str, numeric_re
from frplib.vec_tuples           import VecTuple
from frplib.parsing.parsy_adjust import (generate, parse_error_message,
                                         repeat_within, with_desc, with_label)

# ATTN: Use numeric.Numeric types here for the tree.

#
# Helpers
#

def convert(xs: tuple, f) -> VecTuple:
    return VecTuple(f(x) for x in xs)


#
# Basic Combinators
#

numeric_p = regex(numeric_re).map(numeric_q_from_str)
integer_p = regex(integer_re).map(numeric_q_from_str)

valuei = with_label('a numeric component of a value', numeric_p)
ws = with_label('whitespace', whitespace).optional()
otuple = string('<')
ctuple = string('>')
oparen = string('(')
cparen = string(')')
comma_ws = seq(string(","), ws)

weight = with_label('a branch weight', numeric_p)  # Allow sign so that we can flag the error more clearly
value_tuple = with_desc(
    'a node value',
    seq(ws, otuple, ws) >> valuei.sep_by(comma_ws) << seq(ws, ctuple, ws)
).map(VecTuple)


#
# Main Parser
#

@generate  # ('a kind represented as a string')
def kind_sexp():
    yield oparen
    yield ws
    root = yield value_tuple
    yield ws
    children = yield repeat_within(
        None,
        seq(weight << ws, (kind_sexp | value_tuple) << ws),
        cparen,
        labels={'close': ')', 'item': 'a kind subtree/leaf value'}
    )
    # Need to check dimensions of tuples inside, consistent values, etc.
    result = [root]
    result.extend(children)
    return result

def parse_kind_sexp(s: str, rich=True, short=False) -> list:
    try:
        ktree = kind_sexp.parse(s)
    except ParseError as e:
        raise KindError(parse_error_message(e, rich, short))

    perrors = validate_kind(ktree, sep=", ")
    if perrors:
        raise KindError(f'Errors found in kind string format: {perrors}.')
    return ktree

#
# Validation
#

# Examples:
# [(), [1, [(2,), [3, (2, 4)], [5, (2, 6)]]], [1, [(7,), [8, (7, 9)], [10, (7, 12)]]]]
# [(), [1, (2,)], [3, (4,)]]
# [(), [1, (2, 3)], [4, [(5,), [1, (5, 7)]]]]   # Unusual but OK???
#
# Need:
# 1. All paths consistent:  prefix(node) == parent
# 2. Values added within each subtree are distinct
# 3. All leaves have the same dimension

def check_kind_tree(tree, errors, min_leaf_dim=None, max_leaf_dim=None):
    parent, *branches = tree
    p_str = convert(parent, str)

    def dm(x, v, f):
        if v is None:
            return x
        return f(v, x)

    if len(branches) == 0:
        # Handle root separately, this is an error or just remove these earlier
        return (errors, min_leaf_dim, max_leaf_dim)

    next_added: dict[Any, int] = defaultdict(int)
    for branch in branches:
        weight, subtree = branch
        if weight <= 0:
            errors.append(f'Non-positive weight {weight} at subtree {parent}.')
        if isinstance(subtree, tuple):  # Leaf case
            if len(subtree) <= len(parent):
                st_str = convert(subtree, str)
                errors.append(f'Node {st_str} does not extend {p_str}')
            elif subtree[0:len(parent)] != parent:
                st_str = convert(subtree, str)
                errors.append(f'Node {st_str} inconsistent with parent {p_str}')
            else:
                next_added[subtree[len(parent):]] += 1
                min_leaf_dim = dm(len(subtree), min_leaf_dim, min)
                max_leaf_dim = dm(len(subtree), max_leaf_dim, max)
        else:  # Subtree
            if len(subtree[0]) <= len(parent):
                st_str = convert(subtree[0], str)
                errors.append(f'Node {st_str} does not extend its parent {p_str}')
            elif subtree[0][0:len(parent)] != parent:
                st_str = convert(subtree[0], str)
                errors.append(f'Node {st_str} at subtree inconsistent with parent {p_str}')
            else:
                next_added[subtree[0][len(parent):]] += 1
                errors, min_leaf_dim, max_leaf_dim = check_kind_tree(subtree, errors, min_leaf_dim, max_leaf_dim)

    for k, v in next_added.items():
        if v > 1:
            errors.append(f'Distinct branches have a common prefix '
                          f'<{", ".join([str(x) for x in [*p_str, *k]])}>'
                          f' in subtree directly below {p_str}.')

    return (errors, min_leaf_dim, max_leaf_dim)

def validate_kind(kind_tree, sep="\n"):
    errors, min_dim, max_dim = check_kind_tree(kind_tree, [])
    if min_dim != max_dim:
        errors.append(f'Leaf values in the tree have differing dimensions ({min_dim} < {max_dim}).')
    return sep.join(errors)


#
# Canonical Form
#

def canonical_tree(tree, p=1, weight_fn=as_numeric, value_fn=as_numeric):
    # Assumes tree has been validated already
    parent, *branches = tree

    if len(branches) == 0:
        return [p, convert(parent, value_fn)]

    level_sum = reduce(lambda acc, x: acc + weight_fn(x[0]), branches, 0)
    normalized = [[p * weight_fn(b[0]) / level_sum, b[1]] for b in branches]

    canonical = []
    for branch in normalized:
        weight, subtree = branch
        if isinstance(subtree, tuple):
            canonical.append([weight, convert(subtree, value_fn)])
        else:
            canonical.extend(canonical_tree(subtree, weight, weight_fn, value_fn))
    return canonical

# def canonical_tree(tree, p=1):
#     # Assumes tree has been validated already
#     parent, *branches = tree
#
#     if len(branches) == 0:
#         return [p, parent]
#
#     level_sum = reduce(lambda acc, x: acc + x[0], branches, 0)
#     normalized = [[p * b[0] / level_sum, b[1]] for b in branches]
#
#     canonical = []
#     for branch in normalized:
#         weight, subtree = branch
#         if isinstance(subtree, tuple):
#             canonical.append(branch)
#         else:
#             canonical.extend(canonical_tree(subtree, weight))
#     return canonical



from __future__ import annotations

from parsy     import (
    Parser,
    ParseError,
    Result,
)
from collections.abc   import Iterable
from functools         import wraps
from typing_extensions import Any


#
# Helpers
#

def join_nl(terms: Iterable[str], *, sep: str = ", ", last_sep=', or ', prefixes=['', '', '']) -> str:
    terms = list(terms)
    count = len(terms)
    if count == 1:
        joined = prefixes[0] + terms[0]
    elif count == 2:
        joined = prefixes[1] + f'{terms[0]} or {terms[1]}'
    else:
        joined = prefixes[2] + sep.join(terms[0:-1]) + last_sep + terms[-1]
    return joined


def join_expecteds(expecteds: frozenset[str]):
    return join_nl(list(expecteds), prefixes=['', 'either ', 'one of '])


#
# Fix descriptions and aggregation of results
#

def supplant_result(new_result: Result, base_result: Result | None) -> Result:
    return new_result

def combine_result(new_result: Result, base_result: Result | None) -> Result:
    return new_result.aggregate(base_result)

def extend_result(new_result: Result, base_result: Result | None) -> Result:
    if not base_result:
        return new_result

    if new_result.furthest > base_result.furthest:
        return new_result
    else:
        new_exps = join_expecteds(new_result.expected)
        base_exps = join_expecteds(base_result.expected)
        return Result(
            new_result.status and base_result.status,
            min(new_result.index, base_result.index),
            new_result.value,
            base_result.furthest,
            frozenset([f'{new_exps} after seeing {base_exps}'])
        )

def wrap_result(new_result: Result, base_result: Result | None) -> Result:
    if not base_result:
        return new_result

    if new_result.furthest > base_result.furthest:
        return new_result
    else:
        new_exps = join_expecteds(new_result.expected)
        base_exps = join_expecteds(base_result.expected)
        return Result(
            new_result.status and base_result.status,
            min(new_result.index, base_result.index),
            new_result.value and base_result.value,
            base_result.furthest,
            frozenset([f'{new_exps}, looking for {base_exps}'])
        )

def with_desc(description: str, p: Parser, aggregate=combine_result) -> Parser:
    def p_with_desc(stream, index) -> Result:
        result = p(stream, index)
        if result.status:
            return result
        else:
            return aggregate(Result.failure(index, description), result)
    return Parser(p_with_desc)

def with_label(description: str, p: Parser) -> Parser:
    return with_desc(description, p, aggregate=supplant_result)

# Fixes bug #77 in parsy Parser.desc()
def generate(fn) -> Parser:
    """
    Creates a parser from a generator function
    """
    if isinstance(fn, str):
        return lambda f: with_desc(fn, generate(f))

    @Parser
    @wraps(fn)
    def generated(stream, index):
        # start up the generator
        iterator = fn()

        result = None
        value = None
        try:
            while True:
                next_parser = iterator.send(value)
                result = next_parser(stream, index).aggregate(result)
                if not result.status:
                    return result
                value = result.value
                index = result.index
        except StopIteration as stop:
            returnVal = stop.value
            if isinstance(returnVal, Parser):
                return returnVal(stream, index).aggregate(result)

            return Result.success(index, returnVal).aggregate(result)

    return generated

# Fixes expecteds behavior of builtin Parser method
def bind(parser, bind_fn):
    @Parser
    def bound_parser(stream, index):
        result = parser(stream, index)
        if result.status:
            next_parser = bind_fn(result.value)
            next_result = next_parser(stream, result.index)
            print('>>= ', next_parser, result, next_result)
            return extend_result(next_result, result)
        else:
            return result

    return bound_parser
#
# More functional version of util with better messages
#

def repeat_within(open_delim: Parser | None, item: Parser, close_delim: Parser,
                  min: int = 0, max: int = 1000000000000,  # float("inf"),
                  labels: dict[str, str] = {}) -> Parser:
    """
    ATTN
    Returns a parser that expects the initial parser followed by ``other``.
    The initial parser is expected at least ``min`` times and at most ``max`` times.
    By default, it does not consume ``other`` and it produces a list of the
    results excluding ``other``. If ``consume_other`` is ``True`` then
    ``other`` is consumed and its result is included in the list of results.
    """

    @Parser
    def within_parser(stream, index):
        values: list[Any] = []
        times = 0
        default_labels = {
            'open': 'opening delimiter',
            'close': 'closing delimiter',
            'item': 'item'
        }
        full_labels = default_labels | labels

        if open_delim is not None:
            res = open_delim(stream, index)
            if not res.status:
                Result.failure(index, full_labels['open'])

        while True:

            # try parser first
            res = close_delim(stream, index)
            if res.status and times >= min:
                return Result.success(res.index, values)

            # exceeded max?
            if times >= max:
                # return failure, it matched parser more than max times
                return Result.failure(index, f"{full_labels['item']} at most {max} times")

            # looking for an item
            result = item(stream, index)
            if result.status:
                # consume
                values.append(result.value)
                index = result.index
                times += 1
            elif times >= min:
                # return failure, parser is not followed by other
                return Result.failure(index, full_labels['item']).aggregate(result)  # ATTN
            else:
                # return failure, it did not match parser at least min times
                return Result.failure(index, f"{full_labels['item']} at least {min} times, saw {times}")

    return within_parser


#
# Handling Parsy ParseError
#
# ATTN: Want to handle rich/formatted_text as mentioned here

def parse_error_message(e: ParseError, rich=True, short=False) -> str:  # FormattedText:
    # short implies not rich; i.e., only check if rich is false
    start_context = max(e.index - 5, 0)
    end_context = min(e.index + 6, len(e.stream))
    joined = join_expecteds(e.expected)
    cont = '...' if e.index > 8 else ''
    pad = 3 if e.index > 8 else 0

    # Rich text here: gray for good context, red for bad, and bold red for the pointer
    parsed = cont + e.stream[start_context:e.index]  # style='class:parse.parsed'
    error = e.stream[e.index]  # style='class:parse.error')
    rest = e.stream[(e.index + 1):end_context]
    indent = '     ' + ' ' * (e.index - start_context + pad)
    mark = '^'  # style='class:parse.error')

    mesg = f'I expected to see {joined} at character {e.index + 1}:\n'
    # text = '    "' + parsed + error + e.stream[(e.index + 1):end_context] + '"'
    # pntr = '     ' + (' ' * (e.index - start_context + pad)) + mark
    #
    # return f'{mesg}\n{text}\n{pntr}'

    # return FormattedText([
    #     ('', mesg),
    #     ('', '    "'),
    #     ('class:parse.parsed', parsed),
    #     ('class:parse.error', error),
    #     ('', e.stream[(e.index + 1):end_context] + '"\n'),
    #     ('', '     ' + (' ' * (e.index - start_context + pad))),
    #     ('class:parse.error', mark),
    # ])

    if rich:  # Using rich styles
        return f'{mesg}    "[#71716f]{parsed}[/][#ff0f0f bold]{error}[/]{rest}"\n{indent}[#ff0f0f bold]{mark}[/]'
    elif short:
        return (f'Expected {joined} at character {e.index + 1}: '
                f'"{parsed}*{error}{rest}"')
    else:
        return f'{mesg}    "{parsed}{error}{rest}"\n{indent}{mark}'



from __future__ import annotations

from frplib.kinds      import Kind
from frplib.symbolic   import Symbolic
from frplib.vec_tuples import VecTuple


#
# Kind and FRP based calculations
#

def substitute(quantity, mapping):
    "Substitutes values for symbols in `quantity` from `mapping`."
    if hasattr(quantity, 'this'):  # Facades
        quantity = getattr(quantity, 'this')

    if isinstance(quantity, Symbolic):
        return quantity.substitute(mapping)

    if isinstance(quantity, VecTuple):
        return quantity.map(substitute_with(mapping))

    if isinstance(quantity, Kind):
        f = substitute_with(mapping)
        return quantity.bimap(f, f)

    if isinstance(quantity, list):
        return [substitute(q, mapping) for q in quantity]

    return quantity

def substitute_with(mapping):
    """Returns a function that substitutes values for symbols from `mapping`.

    See `substitute`.

    """
    def sub(quantity, **kw):
        return substitute(quantity, mapping | kw)
    return sub

def substitution(quantity, **kw):
    """Substitutes values for symbols in quantity, with mapping from keywords.

    """
    return substitute(quantity, kw)



from __future__ import annotations

from frplib.symbolic  import symbol, gen_symbol

#
# Helpers
#

def test_symbol_arithmetic():
    "Simple sequence patterns."
    a = symbol('a')
    b = symbol('b')
    c = symbol('c')

    assert str(a) == 'a'
    assert str(b) == 'b'
    assert str(a - a) == '0'
    assert str(1 + a) == '1 + a'
    assert str((1 + a) * (1 - a)) == '1 + -1 a^2'
    assert str(1 + b + b**2 + b**3) == '1 + b + b^2 + b^3'
    assert str((a - a) * (1 + b + b**2 + b**3)) == '0'

    assert str(1.2 * a) == '1.2 a'
    assert str((1.2 * a) ** 2) == '1.44 a^2'

    assert str(gen_symbol()) != str(gen_symbol())

    assert str(1 / (1 + a)) == '1/(1 + a)'
    assert str((1 + a) / a) == '(1 + a)/a'
    assert str((1 + a) / a**2) == '(1 + a)/a^2'
    assert str((1 + a) / (2 * a**2)) == '(1 + a)/(2 a^2)'

    assert str(((1 + a) * (1 - a) * (1 + a**2)) / (1 - a**4)) == '1'
    assert str(a / (1 - a)) == 'a/(1 + -1 a)'
    assert str(1 + a / (1 - a)) == '1/(1 + -1 a)'

    s = a + b + c
    assert str(s) == 'a + b + c'
    assert str(a / s) == 'a/(a + b + c)'
    assert str(a / s + b / s + c / s) == '1'
    assert str(0.5 * a / s) == '0.5 a/(a + b + c)'

    p = a / (1 - a)
    q = 1 / (1 - a)
    assert str(1 / (1 + p)) == '1 + -1 a'
    assert str(a * q) == 'a/(1 + -1 a)'



from __future__ import annotations

import math
import random
import re

from collections.abc   import Collection, Iterable
from dataclasses       import dataclass
from enum              import Enum, auto
from itertools         import chain, combinations, permutations, product
from typing            import Literal, Callable, overload, Union
from typing_extensions import TypeAlias


from rich              import box
from rich.panel        import Panel

from frplib.env        import environment
from frplib.exceptions import ConstructionError, KindError, MismatchedDomain
from frplib.kind_trees import (KindBranch,
                               canonical_from_sexp, canonical_from_tree,
                               unfold_tree, unfolded_labels, unfold_scan, unfolded_str)
from frplib.numeric    import Numeric, ScalarQ, as_numeric
from frplib.protocols  import Projection
from frplib.quantity   import as_quantity, as_quant_vec, as_real_quantity, show_quantities, show_qtuples
from frplib.statistics import Statistic
from frplib.symbolic   import Symbolic, gen_symbol, is_symbolic, symbol
from frplib.utils      import (compose, const, ensure_tuple, identity,
                               is_interactive, is_tuple, lmap, some)
from frplib.vec_tuples import VecTuple, as_numeric_vec, as_vec_tuple, vec_tuple


#
# Types (ATTN)
#


CanonicalKind: TypeAlias = list['KindBranch']
ValueType: TypeAlias = VecTuple[Numeric]  # ATTN


#
# Constants
#


#
# Helpers
#

# ATTN: this should probably become static methods; see Conditional Kinds.
def value_map(f, kind=None):  # ATTN: make in coming maps tuple safe; add dimension hint even if no kind
    # We require that all kinds returned by f are the same dimension
    # But do not check if None is passed explicitly for kind
    if callable(f):
        if kind is not None and len(set([f(ensure_tuple(vs)).dim for vs in kind.value_set])) != 1:
            raise KindError('All values for a transform or mixture must be '
                            'associated with a kind of the same dimension')
        return f
    elif isinstance(f, dict):
        if kind is not None:
            if {ensure_tuple(vs) for vs in f.keys()} < kind.value_set:   # superset of values ok
                raise KindError('All values for the kind must be present in a mixture')
            if len({k.dim for k in f.values()}) != 1:
                raise KindError('All values for a mixture must be associated with a kind of the same dimension')
        scalars = [vs for vs in f.keys() if not is_tuple(vs) and (vs,) not in f]
        if len(scalars) > 0:  # Keep scalar keys but tuplize them as well
            f = f | {(vs,): f[vs] for vs in scalars}  # Note: not mutating on purpose
        return (lambda vs: f[vs])
    # return None
    # move this error to invokation ATTN
    raise KindError('[red]Invalid value transform or mixture provided[/]: '
                    '[italic]should be function or mapping dictionary[/]')

def normalize_branches(canonical) -> list[KindBranch]:
    seen: dict[tuple, KindBranch] = {}
    total = sum(map(lambda b: b.p, canonical))
    for branch in canonical:
        if branch.vs in seen:
            seen[branch.vs] = KindBranch.make(vs=branch.vs, p=seen[branch.vs].p + branch.p / total)
        else:
            seen[branch.vs] = KindBranch.make(vs=branch.vs, p=branch.p / total)
    return sorted(seen.values(), key=lambda b: b.vs)

class EmptyKindDescriptor:
    def __get__(self, obj, objtype=None):
        return objtype([])

class Kind:
    """
    The Kind of a Fixed Random Payoff

    """
    # str | CanonicalKind[a, ProbType] | KindTree[a, ProbType] | Kind[a, ProbType] -> None
    def __init__(self, spec) -> None:
        # branches: CanonicalKind[ValueType, ProbType]
        if isinstance(spec, Kind):
            branches = spec._canonical   # Shared structure OK, Kinds are immutable
        elif isinstance(spec, str):
            branches = canonical_from_sexp(spec)
        elif len(spec) == 0 or isinstance(spec[0], KindBranch):  # CanonicalKind
            branches = normalize_branches(spec)
        else:  # General KindTree
            branches = canonical_from_tree(spec)

        self._canonical: CanonicalKind = branches
        self._size = len(branches)
        self._dimension = 0 if self._size == 0 else len(branches[0].vs)
        self._value_set: set | None = None

    @property
    def size(self):
        "The size of this kind."
        return self._size

    @property
    def dim(self):
        "The dimension of this kind."
        return self._dimension

    def _set_value_set(self):
        elements = []
        for branch in self._canonical:
            elements.append(branch.vs)
        self._value_set = set(elements)

    @property
    def values(self):
        "A user-facing view of the possible values for this kind, with scalar values shown without tuples."
        if self._value_set is None:
            self._set_value_set()   # ensures ._value_set not None
        if self.dim == 1:
            return {x[0] for x in self._value_set}  # type: ignore
        return self._value_set

    @property
    def value_set(self):
        "The raw set of possible values for this kind"
        if self._value_set is None:
            self._set_value_set()
        return self._value_set

    @property
    def _branches(self):
        return self._canonical.__iter__()

    @property
    def weights(self):
        "A dictionary of a kinds canonical weights by value."
        # ATTN: wrap this in a pretty_dict from output.py
        return {b.vs: b.p for b in self._canonical}

    def clone(self):
        "Kinds are immutable, so cloning it just returns itself."
        return self

    # Functorial Methods

    # Note 0: Move to keeping everything as a tuple/VecTuple, show the <> for scalars too, reduce this complexity!
    # Note 1: Remove empty kinds in mixtures
    # Note 2: Can we abstract this into a KindMonad superclass using returns style declaration
    #         Then specialize the types of the superclass to tuple[a,...] and something for probs
    # Note 3: Maybe (following 2) the applicative approach should have single functions at the nodes
    #         rather than tuples (which works), because we can always fork the functions to *produce*
    #         tuples, and then the applicative instance is not dependent on the tuple structure
    # Note 4: We want to allow for other types as the values, like functions or bools or whatever;
    #         having kind monad makes that possible. All clean up to tuple-ify things can happen
    #         *here*.
    # Note 5: Need to allow for synonyms of boolean and 0-1 functions in processing maps versus filterings
    #         so events can be used for both and normal predicates can be used
    # Note 6: Need to work out the proper handling of tuples for these functions. See Statistics object
    #         currently in kind_tests.py.  Add a KindUtilities which defines the constructors, statistics,
    #         and other helpers (fork, join, chain, compose, identity, ...)
    # Note 7: Need to improve initialization and use nicer forms in the utilities below
    # Note 8: Have a displayContext (a default, a current global, and with handlers) that governs
    #         i. how kinds are displayed (full, compact, terse), ii. number system used,
    #         iii. rounding and other details such as whether to reduce probs to lowest form, ...
    #         iv. whether to transform values..... The kind can take a context argument that if not None
    #         overrides the surrounding context in the fields supplied.
    # Note 9: Other things: Add annotations to branches to allow nice modeling. Show event {0,1}
    #         trees as their annotated 1 string if present? Formal linear combinations in expectation when not numeric.
    #         Handling boolean and {0,1} equivalently in predicates (so events are as we describe them later)
    # Note A: Large powers maybe can be handled differently to get better performance; or have a reducing method
    #         when doing things like  d6 ** 10 ^ (Sum / 10 - 5)

    def map(self, f):
        "A functorial transformation of this kind. This is for internal use; use .transform() instead."
        new_kind = lmap(KindBranch.bimap(f), self._canonical)
        return Kind(new_kind)

    def apply(self, fn_kind):  # Kind a -> Kind[a -> b] -> Kind[b]
        "An applicative <*> operation on this kind. (For internal use)"
        def app(branch, fn_branch):
            return [KindBranch.make(vs=f(b), p=branch.p * fn_branch.p) for b in branch.vs for f in fn_branch.vs]
        new_kind = []
        for branch in self._canonical:
            for fn_branch in fn_kind._canonical:
                new_kind.extend(app(branch, fn_branch))
        return Kind(new_kind)

    def bind(self, f):   # self -> (a -> Kind[b, ProbType]) -> Kind[b, ProbType]
        "Monadic bind for this kind. (For internal use)"
        def mix(branch):  # KindBranch[a, ProbType] -> list[KindBranch[b, ProbType]]
            subtree = f(branch.vs)._canonical
            return map(lambda sub_branch: KindBranch.make(vs=sub_branch.vs, p=branch.p * sub_branch.p), subtree)

        new_kind = []
        for branch in self._canonical:
            new_kind.extend(mix(branch))
        return Kind(new_kind)

    def bimap(self, value_fn, weight_fn=identity):
        "A functorial transformation of this kind. This is for internal use; use .transform() instead."
        new_kind = lmap(KindBranch.bimap(value_fn, weight_fn), self._canonical)
        return Kind(new_kind)

    @classmethod
    def unit(cls, value):  # a -> Kind[a, ProbType]
        "Returns the monadic unit for this kind. (For internal use)"
        return Kind([KindBranch.make(as_quant_vec(value), 1)])

    # @classmethod
    # @property
    # def empty( cls ):
    #     return Kind([])
    empty = EmptyKindDescriptor()

    @staticmethod
    def table(kind):
        "DEPRECATED. "
        print(str(kind))

    # Calculations

    def mixture(self, cond_kind):   # Self -> (a -> Kind[a, ProbType]) -> Kind[a, ProbType]
        """Kind Combinator: Creates a mixture kind with this kind as the mixer and `f_mapping` giving the targets.

        This is usually more easily handled by the >> operator, which takes the mixer on the
        left and the target on the right and is equivalent.

        It is recommended that `cond_kind` be a conditional Kind, though this function
        accepts a variety of formats as described below.

        Parameters
        ----------
          cond_kind - either a conditional Kind, a dictionary taking values of this
                      kind to other kinds, or a function doing the same. Every possible
                      value of this kind must be represented in the mapping. For scalar
                      kinds, the values in the dictionary or function can be scalars,
                      as they will be converted to the right form in this function.

        Returns a new mixture kind that combines the mixer and targets.

        """
        f = value_map(cond_kind, self)

        def join_values(vs):
            new_tree = f(vs)._canonical
            if len(new_tree) == 0:      # Empty result tree  (ATTN:CHECK)
                new_tree = [KindBranch.make(vs=(), p=1)]
            return Kind([KindBranch.make(vs=tuple(list(vs) + list(branch.vs)), p=branch.p) for branch in new_tree])

        return self.bind(join_values)

    def independent_mixture(self, kind_spec):
        """Kind Combinator: An independent mixture of this kind with another kind.

        This is usually more easily handled by the * operator, which is equivalent.

        Parameter `kind_spec` should be typically be a valid kind,
        but this will accept anything that produces a valid kind via
        the `kind()` function.

        Returns a new kind representing this mixture.

        """
        r_kind = kind(kind_spec)

        if len(r_kind) == 0:
            return self
        if len(self) == 0:
            return r_kind

        def combine_product(branchA, branchB):
            return KindBranch.make(vs=list(branchA.vs) + list(branchB.vs), p=branchA.p * branchB.p)

        return Kind([combine_product(brA, brB) for brA, brB in product(self._canonical, r_kind._canonical)])

    def transform(self, statistic):
        """Kind Combinator: Transforms this kind by a statistic. Returns a new kind.

        This is often more easily handled by the ^ operator, or by direct composition
        by the statistic, which are equivalent.

        """
        # Handle statistics case carefully; this all needs a retune
        # Check wrapping, dims/codims etc.
        # This is temporary
        if isinstance(statistic, Statistic):
            if statistic.dim == 0 or statistic.dim == self.dim or self.dim == 0:
                f = statistic
            else:
                try:
                    statistic(self._canonical[0].vs)
                    f = statistic
                except Exception:
                    raise KindError(f'Statistic {statistic.name} is incompatible with this kind.')
        else:
            f = compose(ensure_tuple, value_map(statistic))  # ATTN!
        return self.map(f)

    def conditioned_on(self, cond_kind):
        """Kind Combinator: computes the kind of the target conditioned on the mixer (this kind).

        This is usually more clearly handled with the // operator, which takes mixer // target.

        This is related to, but distinct from, a mixture in that it produces the kind
        of the target, marginalizing out this kind. Conditioning is the operation of
        using hypothetical information about one kind and a contingent relationship between
        them to compute another kind.
        """
        if cond_kind is None:  # ATTN: None not a possible return value of value_map  at the moment
            raise KindError('Conditioning on this kind requires a valid and '
                            'matching mapping of values to kinds of the same dimension')
        return self.bind(value_map(cond_kind, self))

    def expectation(self):
        """Computes the expectation of this kind. Scalar expectations are unwrapped. (Internal use.)

        The expectation should be computed using the E operator rather than this method.
        """
        ex = [as_numeric(0)] * self.dim
        for branch in self._canonical:
            for i, v in enumerate(branch.vs):
                ex[i] += branch.p * v
        return ex[0] if self.dim == 1 else as_vec_tuple(ex)

    # Overloads

    def __eq__(self, other) -> bool:
        if not isinstance(other, Kind):
            return False
        return self._canonical == other._canonical

    def __len__(self) -> int:
        return self._size

    def __bool__(self) -> bool:
        return self._dimension > 0

    def __iter__(self):
        yield from ((b.p, b.vs) for b in self._canonical)

    def __mul__(self, other):
        "Mixes FRP with another independently"
        return self.independent_mixture(other)

    def __pow__(self, n, modulo=None):
        "Mixes FRP with itself n times independently"
        # Use monoidal power trick
        if n < 0:
            raise KindError('Kind powers with negative exponents not allowed')
        elif n == 0 or self.dim == 0:
            return Kind.empty
        elif n == 1:
            return self

        def combine_product(orig_branches):
            vs = []
            p = 1
            for b in orig_branches:
                vs.extend(b.vs)
                p *= b.p
            return KindBranch.make(vs=vs, p=p)

        return Kind([combine_product(obranches) for obranches in product(self._canonical, repeat=n)])

    def __rfloordiv__(self, other):
        "Conditioning on self; other is a conditional distribution."
        conditional_dist = value_map(other, self)
        if conditional_dist is None:  # ATTN: None no longer returned, think about managing this
            return NotImplemented
        return self.bind(conditional_dist)

    def __rshift__(self, f_mapping):
        "Mixes FRP with FRPs given for each value"
        return self.mixture(f_mapping)

    def __xor__(self, f_mapping):
        "Applies a transform to an FRP"
        return self.transform(f_mapping)

    # Need a protocol for ProjectionStatistic to satisfy to avoid circularity
    @overload
    def marginal(self, *__indices: int) -> 'Kind':
        ...

    @overload
    def marginal(self, __subspace: Iterable[int] | Projection | slice) -> 'Kind':
        ...

    def marginal(self, *index_spec) -> 'Kind':
        dim = self.dim

        # Unify inputs
        if len(index_spec) == 0:
            return Kind.empty
        if isinstance(index_spec[0], Iterable):
            indices: tuple[int, ...] = tuple(index_spec[0])
        elif isinstance(index_spec[0], Projection):
            indices = tuple(index_spec[0].subspace)
        elif isinstance(index_spec[0], slice):
            start, stop, step = index_spec[0].indices(dim + 1)
            indices = tuple(range(start, stop, step))
        else:
            indices = index_spec

        # Check dimensions (allow negative indices python style)
        if any([index == 0 or index < -dim or index > dim for index in indices]):
            raise KindError( f'All marginalization indices in {indices} should be between 1..{dim} or -{dim}..-1')

        # Marginalize
        def marginalize(value):
            return tuple(map(lambda i: value[i - 1] if i > 0 else value[i], indices))
        return self.map(marginalize)

    def __getitem__(self, indices):
        "Marginalizing this kind; other is a projection index or list of indices (1-indexed)"
        try:
            return self.marginal(indices)
        except Exception:
            return NotImplemented

    def __or__(self, predicate):  # Self -> ValueMap[ValueType, bool] -> Kind[ValueType, ProbType]
        "Applies a conditional filter to FRP"
        keep = value_map(predicate)
        return Kind([branch for branch in self._canonical if keep(branch.vs)])

    def sample1(self):
        "Returns the value of one FRP with this kind."
        return VecTuple(self.sample(1)[0])

    def sample(self, n: int = 1):
        "Returns a list of values corresponding to `n` FRPs with this kind."
        weights = [float(branch.p) for branch in self._canonical] or [1]
        values = [branch.vs for branch in self._canonical] or [vec_tuple()]
        # ATTN: Conver to iterator ??
        return lmap(VecTuple, random.choices(values, weights, k=n))

    def show_full(self) -> str:
        """Show a full ascii version of this kind as a tree in canonical form."""
        if len(self._canonical) == 0:
            return '<> -+'

        size = self.size
        juncture, extra = (size // 2, size % 2 == 0)

        p_labels = show_quantities(branch.p  for branch in self._canonical)
        v_labels = show_qtuples(branch.vs for branch in self._canonical)
        pwidth = max(map(len, p_labels), default=0) + 2

        lines = []
        if size == 1:
            plab = ' ' + p_labels[0] + ' '
            vlab = v_labels[0].replace(', -', ',-')  # ATTN:HACK fix elsewhere, e.g., '{0:-< }'.format(Decimal(-16.23))
            lines.append(f'<> ------{plab:-<{pwidth}}---- {vlab}')
        else:
            for i in range(size):
                plab = ' ' + p_labels[i] + ' '
                vlab = v_labels[i].replace(', -', ',-')   # ATTN:HACK fix elsewhere
                if i == 0:
                    lines.append(f'    ,----{plab:-<{pwidth}}---- {vlab}')
                    if size == 2:
                        lines.append('<> -|')
                        # lines.extend(['    |', '<> -|', '    |'])
                elif i == size - 1:
                    lines.append(f'    `----{plab:-<{pwidth}}---- {vlab}')
                elif i == juncture:
                    if extra:
                        lines.append( '<> -|')
                        lines.append(f'    |----{plab:-<{pwidth}}---- {vlab}')
                    else:
                        lines.append(f'<> -+----{plab:-<{pwidth}}---- {vlab}')
                else:
                    lines.append(f'    |----{plab:-<{pwidth}}---- {vlab}')
        return '\n'.join(lines)

    def __str__(self) -> str:
        return self.show_full()

    def __frplib_repr__(self):
        if environment.ascii_only:
            return str(self)
        return Panel(str(self), expand=False, box=box.SQUARE)

    def __repr__(self) -> str:
        if is_interactive():   # ATTN: Do we want this anymore??
            return self.show_full()  # So it looks nice at the repl
        return super().__repr__()

    def repr_internal(self) -> str:
        return f'Kind({repr(self._canonical)})'

# Utilities

@dataclass(frozen=True)
class UnfoldedKind:
    unfolded: list  # KindTree
    upicture: str

    def __str__(self) -> str:
        return self.upicture

    def __repr__(self) -> str:
        return repr(self.unfolded)

    def __frplib_repr__(self):
        if environment.ascii_only:
            return str(self)
        return Panel(str(self), expand=False, box=box.SQUARE)

def unfold(k: Kind) -> UnfoldedKind:  # ATTN: Return an object that prints this string, later
    dim = k.dim
    unfolded = unfold_tree(k._canonical)
    if unfolded is None:
        return UnfoldedKind(k._canonical, k.show_full())
    # ATTN: Remove other components from this, no longer needed

    wd = [(0, 3)]  # Widths of the root node weight (empty) and value (<>)
    labelled = unfolded_labels(unfolded[1:], str(unfolded[0]), 1, wd)
    sep = [2 * (dim - level) for level in range(dim + 1)]  # seps should be even
    scan, _ = unfold_scan(labelled, wd, sep)

    return UnfoldedKind(unfolded, unfolded_str(scan, wd))


# Sequence argument interface

class Flatten(Enum):
    NOTHING = auto()
    NON_TUPLES = auto()
    NON_VECTORS = auto()
    EVERYTHING = auto()

flatteners: dict[Flatten, Callable] = {
    Flatten.NON_TUPLES: lambda x: x if isinstance(x, Iterable) and not isinstance(x, tuple) else [x],
    Flatten.NON_VECTORS: lambda x: x if isinstance(x, Iterable) and not isinstance(x, VecTuple) else [x],
    Flatten.EVERYTHING: lambda x: x if isinstance(x, Iterable) else [x],
}

ELLIPSIS_MAX_LENGTH: int = 10 ** 6


def sequence_of_values(
        *xs: Numeric | Symbolic | Iterable[Numeric | Symbolic] | Literal[Ellipsis],   # type: ignore
        flatten: Flatten = Flatten.NON_VECTORS,
        transform=identity,
        pre_transform=identity,
        parent=''
) -> list[Numeric | Symbolic]:
    # interface that reads values in various forms
    # individual values  1, 2, 3, 4
    # elided sequences   1, 2, ..., 10
    # iterables          [1, 2, 3, 4]
    # mixed sequences    1, 2, [1, 2, 3], 4, range(100,110), (17, 18)   with flatten=True only
    if flatten != Flatten.NOTHING:
        proto_values = list(chain.from_iterable(map(flatteners[flatten], map(pre_transform, xs))))
    elif len(xs) == 1 and isinstance(xs[0], Iterable):
        proto_values = list(pre_transform(xs[0]))
    else:
        proto_values = list(map(pre_transform, xs))

    values = []
    n = len(proto_values)
    for i in range(n):
        value = proto_values[i]
        if value == Ellipsis:
            if i <= 1 or i == n - 1:
                raise KindError(f'Argument ... to {parent or "a factory"} must be appear in the pattern a, b, ..., c.')
            a, b, c = (proto_values[i - 2], proto_values[i - 1], proto_values[i + 1])
            if some(is_symbolic, [a, b, c]):
                raise ConstructionError('An ellipsis ... cannot be used between symbolic quantities')
            if (a - b) * (b - c) <= 0:
                raise KindError(f'Argument ... to {parent or "a factory"} must be appear in the pattern a, b, ..., c '
                                f'with a < b < c or a > b > c.')
            if (c - b) > (b - a) * ELLIPSIS_MAX_LENGTH:
                raise KindError(f'Argument ... to {parent or "a factory"} will leads to a very large sequence;'
                                f"I'm guessing this is a mistake.")
            values.extend([transform(b + k * (b - a)) for k in range(1, int(math.floor((c - b) / (b - a))))])
        else:
            values.append(transform(value))

    return values


#
# Kind Builders
#

void: Kind = Kind.empty

def constant(a) -> Kind:
    """Kind Factory: returns the kind of a constant FRP with the specified value.

    Parameters:
      `a`: any numeric value

    Returns the kind <> --- <a>.

    """
    return Kind.unit(a)

def either(a, b, weight_ratio=1) -> Kind:
    """A choice between two possibilities a and b with ratio of weights (a to b) of `weight_ratio`.

    Values can be numbers, symbols, or strings. In the latter case they are converted
    to numeric or symbolic values as appropriate. Rational values in strings (e.g., '1/7')
    are allowed but must have no space around the '/'.

    """
    ratio = as_numeric(weight_ratio)
    p_a = ratio / (1 + ratio)
    return Kind([KindBranch.make(vs=as_quant_vec(a), p=p_a),
                 KindBranch.make(vs=as_quant_vec(b), p=1 - p_a)])

def uniform(*xs: Numeric | Symbolic | Iterable[Numeric | Symbolic] | Literal[Ellipsis]) -> Kind:   # type: ignore
    """Returns a kind with equal weights on the given values.

    Values can be specified in a variety of ways:
      + As explicit arguments, e.g.,  uniform(1, 2, 3, 4)
      + As an implied sequence, e.g., uniform(1, 2, ..., 10)
        Here, two *numeric* values must be supplied before the ellipsis and one after;
        the former determine the start and increment; the latter the end point.
        Multiple implied sequences with different increments are allowed,
        e.g., uniform(1, 2, ..., 10, 12, ... 20)
      + As an iterable, e.g., uniform([1, 10, 20]) or uniform(irange(1,52))
      + With a combination of methods, e.g.,
           uniform(1, 2, [4, 3, 5], 10, 12, ..., 16)
        in which case all the values except explicit *tuples* will be
        flattened into a sequence of values. (Though note: all values
        should have the same dimension.)

    Values can be numbers, tuples, symbols, or strings. In the latter case they
    are converted to numbers or symbols as appropriate.

    """
    values = sequence_of_values(*xs, flatten=Flatten.NON_TUPLES, transform=as_quant_vec)
    if len(values) == 0:
        return Kind.empty
    return Kind([KindBranch.make(vs=x, p=1) for x in values])

def symmetric(*xs, around=None, weight_by=lambda dist: 1 / dist if dist > 0 else 1) -> Kind:
    """Returns a kind with the given values and weights a symmetric function of the values.

    Specifically, the weights are determined by the distance of each value
    from a specified value `around`:

           weight(x) = weight_fn(distance(x, around))

    If `around` is specified it is used; otherwise, the mean of the values is used.
    The `weight_fn` defaults to 1/distance, but can specified; it should be a function
    of one numeric parameter.

    Values can be specified in a variety of ways:
      + As explicit arguments, e.g.,  symmetric(1, 2, 3, 4)
      + As an implied sequence, e.g., symmetric(1, 2, ..., 10)
        Here, two *numeric* values must be supplied before the ellipsis and one after;
        the former determine the start and increment; the latter the end point.
        Multiple implied sequences with different increments are allowed,
        e.g., symmetric(1, 2, ..., 10, 12, ... 20)
      + As an iterable, e.g., symmetric([1, 10, 20]) or symmetric(irange(1,52))
      + With a combination of methods, e.g.,
           symmetric(1, 2, [4, 3, 5], 10, 12, ..., 16)
        in which case all the values except explicit *tuples* will be
        flattened into a sequence of values. (Though note: all values
        should have the same dimension.)

    Values can be numbers, tuples, or strings. In the latter case they
    are converted to numbers. If values are tuples, then either `around`
    should also be a tuple, or if that is not supplied, the tuples should
    first be passed to the qvec() function to make the distance computable.

    """
    if isinstance(around, tuple):  # We expect tuple values as well
        pre = as_numeric_vec
        post = identity
    else:
        pre = identity
        post = as_numeric_vec
    values = sequence_of_values(*xs, flatten=Flatten.NON_TUPLES, transform=pre)
    n = len(values)
    if n == 0:
        return Kind.empty
    if around is None:
        around = sum(values) / n  # type: ignore
    return Kind([KindBranch.make(vs=post(x), p=as_numeric(weight_by(abs(x - around)))) for x in values])

def linear(
        *xs: Numeric | Symbolic | Iterable[Numeric | Symbolic] | Literal[Ellipsis],  # type: ignore
        first=1,
        increment=1
) -> Kind:
    """Returns a kind with the specified values and weights varying linearly

    Parameters
    ----------
    *xs: The values, see below.
    first: The weight associated with the first value. (Default: 1)
    increment: The increase in weight associated with each
        subsequent value. (Default: 1)

    Values can be specified in a variety of ways:
      + As explicit arguments, e.g.,  linear(1, 2, 3, 4)
      + As an implied sequence, e.g., linear(1, 2, ..., 10)
        Here, two *numeric* values must be supplied before the ellipsis and one after;
        the former determine the start and increment; the latter the end point.
        Multiple implied sequences with different increments are allowed,
        e.g., linear(1, 2, ..., 10, 12, ... 20)
      + As an iterable, e.g., linear([1, 10, 20]) or linear(irange(1,52))
      + With a combination of methods, e.g.,
           linear(1, 2, [4, 3, 5], 10, 12, ..., 16)
        in which case all the values except explicit *tuples* will be
        flattened into a sequence of values. (Though note: all values
        should have the same dimension.)

    Values can be numbers, tuples, symbols, or strings. In the latter case they
    are converted to numbers or symbols as appropriate.
    """
    values = sequence_of_values(*xs, flatten=Flatten.NON_TUPLES, transform=as_quant_vec)
    weights = [as_quantity(first + k * increment) for k in range(len(values))]

    return Kind([KindBranch.make(vs=x, p=w) for x, w in zip(values, weights)])

def geometric(
        *xs: Numeric | Symbolic | Iterable[Numeric | Symbolic] | Literal[Ellipsis],  # type: ignore
        first=1,
        r=0.5
) -> Kind:
    """Returns a kind with the specified values and weights varying geometrically

    Parameters
    ----------
    *xs: The values, see below.
    first: The weight associated with the first value. (Default: 1)
    r: The the ratio between a weight and the preceding weight. (Default: 0.5)

    Values can be specified in a variety of ways:
      + As explicit arguments, e.g.,  geometric(1, 2, 3, 4)
      + As an implied sequence, e.g., geometric(1, 2, ..., 10)
        Here, two *numeric* values must be supplied before the ellipsis and one after;
        the former determine the start and increment; the latter the end point.
        Multiple implied sequences with different increments are allowed,
        e.g., geometric(1, 2, ..., 10, 12, ... 20)
      + As an iterable, e.g., geometric([1, 10, 20]) or geometric(irange(1,52))
      + With a combination of methods, e.g.,
           geometric(1, 2, [4, 3, 5], 10, 12, ..., 16)
        in which case all the values except explicit *tuples* will be
        flattened into a sequence of values. (Though note: all values
        should have the same dimension.)

    Values can be numbers, tuples, symbols, or strings. In the latter case they
    are converted to numbers or symbols as appropriate.
    """
    values = sequence_of_values(*xs, flatten=Flatten.NON_TUPLES, transform=as_quant_vec)
    ratio = as_quantity(r)
    w = as_quantity(first)
    weights = []
    for _ in values:
        weights.append(w)
        w = w * ratio      # type: ignore
    return Kind([KindBranch.make(vs=x, p=w) for x, w in zip(values, weights)])

def weighted_by(*xs, weight_by: Callable) -> Kind:
    """Returns a kind with the specified values weighted by a function

    Parameters
    ----------
    *xs: The values, see below.
    weight_by: A function that takes a value and returns a corresponding weight.

    Values can be specified in a variety of ways:
      + As explicit arguments, e.g.,  weighted_by(1, 2, 3, 4)
      + As an implied sequence, e.g., weighted_by(1, 2, ..., 10)
        Here, two *numeric* values must be supplied before the ellipsis and one after;
        the former determine the start and increment; the latter the end point.
        Multiple implied sequences with different increments are allowed,
        e.g., weighted_by(1, 2, ..., 10, 12, ... 20)
      + As an iterable, e.g., weighted_by([1, 10, 20]) or weighted_by(irange(1,52))
      + With a combination of methods, e.g.,
           weighted_by(1, 2, [4, 3, 5], 10, 12, ..., 16)
        in which case all the values except explicit *tuples* will be
        flattened into a sequence of values. (Though note: all values
        should have the same dimension.)


    Values and weights can be numbers, tuples, symbols, or strings.
    In the latter case they are converted to numbers or symbols as
    appropriate. `weight_by` must return a valid weight for all
    specified values.

    """
    values = sequence_of_values(*xs, flatten=Flatten.NON_TUPLES)
    if len(values) == 0:
        return Kind.empty
    return Kind([KindBranch.make(vs=as_quant_vec(x), p=as_quantity(weight_by(x))) for x in values])

def weighted_as(*xs, weights: list[ScalarQ | Symbolic] = []) -> Kind:
    """Returns a kind with the specified values weighted by given weights.

    Parameters
    ----------
    *xs: The values, see below.
    weights: A list of weights, one per given value. This c
        can include

    Values (and weights) can be specified in a variety of ways:
      + As explicit arguments, e.g.,  weighted_as(1, 2, 3, 4)
      + As an implied sequence, e.g., weighted_as(1, 2, ..., 10)
        Here, two *numeric* values must be supplied before the ellipsis and one after;
        the former determine the start and increment; the latter the end point.
        Multiple implied sequences with different increments are allowed,
        e.g., weighted_as(1, 2, ..., 10, 12, ... 20)
      + As an iterable, e.g., weighted_as([1, 10, 20]) or weighted_as(irange(1,52))
      + With a combination of methods, e.g.,
           weighted_as(1, 2, [4, 3, 5], 10, 12, ..., 16)
        in which case all the values except explicit *tuples* will be
        flattened into a sequence of values. (Though note: all values
        should have the same dimension.)

    Values and weights can be numbers, tuples, symbols, or strings.
    In the latter case they are converted to numbers or symbols as
    appropriate.

    """
    values = sequence_of_values(*xs, flatten=Flatten.NON_TUPLES)
    if len(values) == 0:
        return Kind.empty

    kweights: list[Union[Numeric, Symbolic]] = sequence_of_values(*weights, flatten=Flatten.NON_TUPLES)
    if len(kweights) < len(values):
        kweights = [*kweights, *([1] * (len(values) - len(kweights)))]

    return Kind([KindBranch.make(vs=as_quant_vec(x), p=as_quantity(w))
                 for x, w in zip(values, kweights)])

def arbitrary(*xs, names: list[str] = []):
    "Returns a kind with the given values and arbitrary symbolic weights."
    values = sequence_of_values(*xs, flatten=Flatten.NON_TUPLES, transform=as_numeric_vec)
    if len(values) == 0:
        return Kind.empty
    syms = lmap(symbol, names)
    for i in range(len(values) - len(syms)):
        syms.append(gen_symbol())
    return Kind([KindBranch.make(vs=x, p=sym) for x, sym in zip(values, syms)])

def integers(start, stop=None, step: int = 1, weight_fn=lambda _: 1):
    if stop is None:
        stop = start
        start = 0
    if (stop - start) * step <= 0:
        return Kind.empty
    return Kind([KindBranch.make(vs=as_numeric_vec(x), p=weight_fn(x)) for x in range(start, stop, step)])

def evenly_spaced(start, stop=None, num: int = 2, weight_by=lambda _: 1):
    if stop is None:
        stop = start
        start = 0
    if math.isclose(start, stop) or num < 1:
        return Kind.empty
    if num == 1:
        return Kind.unit(start)
    step = abs(start - stop) / (num - 1)
    return Kind([KindBranch.make(vs=(x,), p=weight_by(x))
                 for i in range(num) if (x := start + i * step) is not None])

def without_replacement(n: int, xs: Iterable) -> Kind:
    "Kind of an FRP that samples n items from a set without replacement."
    return Kind([KindBranch.make(vs=comb, p=1) for comb in combinations(xs, n)])

def subsets(xs: Collection) -> Kind:
    return without_replacement(len(xs), xs)

def permutations_of(xs: Collection, r=None) -> Kind:
    return Kind([KindBranch.make(vs=pi, p=1) for pi in permutations(xs, r)])

# ATTN: lower does not need to be lower just any bin boundary (but watch the floor below)
def bin(scalar_kind, lower, width):
    ""
    if scalar_kind.dim > 1:
        raise KindError(f'Binning of non-scalar kinds (here of dimension {scalar_kind.dim} not yet supported')
    values: dict[tuple, Numeric] = {}
    for branch in scalar_kind._canonical:
        bin = ( lower + width * math.floor((branch.value - lower) / width), )
        if bin in values:
            values[bin] += branch.p
        else:
            values[bin] = branch.p
    return Kind([KindBranch.make(vs=v, p=p) for v, p in values.items()])


#
# Utilities
#
# See also the generic utilities size, dim, values, frp, unfold, clone, et cetera.

def kind(any) -> Kind:
    "A generic constructor for kinds, from strings, other kinds, FRPs, and more."
    if isinstance(any, Kind):
        return any
    if hasattr(any, 'kind'):
        return any.kind
    if not any:
        return Kind.empty
    if isinstance(any, str) and (any in {'void', 'empty'} or re.match(r'\s*\(\s*<\s*>\s*\)\s*', any)):
        return Kind.empty
    try:
        return Kind(any)
    except Exception as e:
        raise KindError(f'I could not create a kind from {any}: {str(e)}')


#
# Conditional Kinds
#

class ConditionalKind:
    """A unified representation of a conditional Kind.

    A conditional Kind is a mapping from a set of values of common
    dimension to Kinds of common dimension. This can be based
    on either a dictionary or on a function, though the dictionary
    is often more convenient in practice as we can determine the
    domain easily.

    This provides a number of facilities that are more powerful than
    working with a raw dictionary or function: nice output at the repl,
    automatic conversion of values, and automatic expectation computation
    (as a function from values to predictions). It is also more robust
    as this conversion performs checks and corrections.

    To create a conditional kind, use the `conditional_kind` function,
    which see.

    """
    def __init__(
            self,
            mapping: Callable[[ValueType], Kind] | dict[ValueType, Kind] | Kind,
            *,
            codim: int | None = None,
            dim: int | None = None,
            domain: Iterable[ValueType] | None = None
    ) -> None:
        # These are optional hints, useful for checking compatibility
        self._codim = codim
        self._dim = dim
        self._domain = set(domain) if domain is not None else None
        self._is_dict = True
        self._original_fn: Callable[[ValueType], Kind] | None = None

        if isinstance(mapping, Kind):
            mapping = const(mapping)

        if isinstance(mapping, dict):
            self._mapping: dict[ValueType, Kind] = {as_numeric_vec(k): v for k, v in mapping.items()}

            def fn(*args) -> Kind:
                if len(args) == 0:
                    raise MismatchedDomain('A conditional Kind requires an argument, none were passed.')
                if isinstance(args[0], tuple):
                    if self._codim and len(args[0]) != self._codim:
                        raise MismatchedDomain(f'A value of dimension {len(args[0])} passed to a'
                                               f' conditional Kind of mismatched codim {self._codim}.')
                    value = as_numeric_vec(args[0])   # ATTN: VecTuple better here?
                elif self._codim and len(args) != self._codim:
                    raise MismatchedDomain(f'A value of dimension {len(args)} passed to a '
                                           f'conditional Kind of mismatched codim {self._codim}.')
                else:
                    value = as_numeric_vec(args)
                if value not in self._mapping:
                    raise MismatchedDomain(f'Value {value} not in domain of conditional Kind.')
                return self._mapping[value]

            self._fn: Callable[..., Kind] = fn

            if self._dim and any(v.dim != self._dim for _, v in self._mapping.items()):
                raise ConstructionError('The Kinds produced by a conditional Kind are not all of the same dimension')
        elif callable(mapping):         # Check to please mypy
            self._mapping = {}
            self._is_dict = False
            self._original_fn = mapping

            def fn(*args) -> Kind:
                if len(args) == 0:
                    raise MismatchedDomain('A conditional Kind requires an argument, none were passed.')
                if isinstance(args[0], tuple):
                    if self._codim and len(args[0]) != self._codim:
                        raise MismatchedDomain(f'A value of dimension {len(args[0])} passed to a'
                                               f' conditional Kind of mismatched codim {self._codim}.')
                    value = as_numeric_vec(args[0])
                elif self._codim and len(args) != self._codim:
                    raise MismatchedDomain(f'A value of dimension {len(args)} passed to a '
                                           f'conditional Kind of mismatched codim {self._codim}.')
                else:
                    value = as_numeric_vec(args)
                if self._domain and value not in self._domain:
                    raise MismatchedDomain(f'Value {value} not in domain of conditional Kind.')

                if value in self._mapping:
                    return self._mapping[value]
                try:
                    result = mapping(value)
                except Exception as e:
                    raise MismatchedDomain(f'encountered a problem passing {value} to a conditional Kind: {str(e)}')
                self._mapping[value] = result   # Cache, fn shold be pure
                return result

            self._fn = fn

    def __call__(self, *value) -> Kind:
        return self._fn(*value)

    def __getitem__(self, *value) -> Kind:
        return self._fn(*value)

    def clone(self) -> 'ConditionalKind':
        "Returns a clone of this conditional kind, which being immutable is itself."
        return self

    def map(self, transform) -> dict | Callable:
        "Returns a dictionary or function like this conditional kind applying `transform` to each kind."
        if self._is_dict:
            return {k: transform(v) for k, v in self._mapping.items()}

        fn = self._original_fn
        assert callable(fn)

        def trans_map(*x):
            return transform(fn(*x))
        return trans_map

    def expectation(self) -> Callable:
        """Returns a function from values to the expectation of the corresponding kind.

        The domain, dim, and codim of the conditional kind are each included as an
        attribute ('domain', 'dim', and 'codim', respetively) of the returned
        function. These may be None if not available.

        """
        def fn(*x):
            try:
                k = self._fn(*x)
            except MismatchedDomain:
                return None
            return k.expectation()

        setattr(fn, 'codim', self._codim)
        setattr(fn, 'dim', self._dim)
        setattr(fn, 'domain', self._domain)

        return fn

    def __str__(self) -> str:
        pad = ': '
        tbl = '\n\n'.join([show_labeled(self._mapping[k], str(k) + pad) for k in self._mapping])
        label = ''
        dlabel = ''
        if self._codim:
            label = label + f' from values of dimension {str(self._codim)}'
        if self._dim:
            label = label + f' to values of dimension {str(self._dim)}'
        if self._domain:
            dlabel = f' with domain={str(self._domain)}'

        if self._is_dict or self._domain == set(self._mapping.keys()):
            title = 'A conditional Kind with mapping:\n'
            return title + tbl
        elif tbl:
            mlabel = f'\nIt\'s mapping includes:\n{tbl}\n  ...more kinds\n'
            return f'A conditional Kind as a function{dlabel or label or mlabel}'
        return f'A conditional Kind as a function{dlabel or label}'

    def __frplib_repr__(self):
        if environment.ascii_only:
            return str(self)
        return Panel(str(self), expand=False, box=box.SQUARE)

    def __repr__(self) -> str:
        if environment.is_interactive:
            return str(self)
        label = ''
        if self._codim:
            label = label + f', codim={repr(self._codim)}'
        if self._dim:
            label = label + f', dim={repr(self._dim)}'
        if self._domain:
            label = label + f', domain={repr(self._domain)}'
        if self._is_dict or self._domain == set(self._mapping.keys()):
            return f'ConditionalKind({repr(self._mapping)}{label})'
        else:
            return f'ConditionalKind({repr(self._fn)}{label})'


def conditional_kind(
        mapping: Callable[[ValueType], Kind] | dict[ValueType, Kind] | Kind | None = None,
        *,
        codim: int | None = None,
        dim: int | None = None,
        domain: set | None = None
) -> ConditionalKind | Callable[..., ConditionalKind]:
    """Converts a mapping from values to FRPs into a conditional FRP.

    While an arbitrary mapping can be used ATTN

    """
    if mapping is not None:
        return ConditionalKind(mapping, codim=codim, dim=dim, domain=domain)

    def decorator(fn: Callable) -> ConditionalKind:
        return ConditionalKind(fn, codim=codim, dim=dim, domain=domain)
    return decorator


#
# Provisional for incorporation and testing
#

def show_labeled(kind, label, width=None):
    width = width or len(label) + 1
    label = f'{label:{width}}'
    return re.sub(r'^.*$', lambda m: label + m[0] if re.match(r'\s*<>', m[0]) else (' ' * width) + m[0],
                  str(kind), flags=re.MULTILINE)


def tbl(mix, pad=': '):
    print( '\n\n'.join([show_labeled(mix[k], str(k) + pad) for k in mix]))
